#
# -*- coding: utf-8 -*-
from __future__ import print_function
#
#
def pca_star(**kwargs):
    """Wrapper on pca_solve to handle stellar eigenspectra.
    """
    import os
    import os.path
    import pickle
    import pylab
    from astropy.io import fits as pyfits
    import numpy as np
    import pydl.pydlutils.yanny as yanny
    from matplotlib.font_manager import fontManager, FontProperties
    from ... import uniq
    from ...pydlutils.goddard.astro import get_juldate
    from ...pydlutils.image import djs_maskinterp
    from . import pca_solve, readspec, skymask
    if 'inputfile' in kwargs:
        inputfile = kwargs['inputfile']
    else:
        inputfile = os.path.join(os.getenv('IDLSPEC2D_DIR'),
            'templates','eigeninput_star.par')
    wavemin = 0
    wavemax = 0
    snmax = 100.0
    if 'niter' in kwargs:
        niter = kwargs['niter']
    else:
        niter = 10
    cspeed = 2.99792458e5
    #
    # Name the output files.
    #
    jd = get_juldate()
    outfile = "spEigenStar-{0:d}".format(int(jd - 2400000.5))
    #
    # Read the input spectra
    #
    par = yanny.yanny(inputfile,np=True)
    slist = par['EIGENOBJ']
    spplate = readspec(slist['plate'],slist['fiberid'],mjd=slist['mjd'],align=True,**kwargs)
    objdloglam = spplate['loglam'][0,1] - spplate['loglam'][0,0]
    #
    # Insist that all of the requested spectra exist.
    #
    missing = spplate['plugmap']['FIBERID'] == 0
    if missing.any():
        imissing = missing.nonzero()[0]
        for k in imissing:
            print("Missing plate={0:d} mjd={1:d} fiber={2:d}".format(plate[k],mjd[k],fiber[k]))
        raise ValueError("{0:d} missing object(s).".format(missing.sum()))
    #
    # Do not fit where the spectrum may be dominated by sky-sub residuals.
    #
    objinvvar = skymask(spplate['invvar'],spplate['andmask'],spplate['ormask'])
    ifix = spplate['flux']**2 * objinvvar > snmax**2
    if ifix.any():
        objinvvar[ifix.nonzero()] = (snmax/spplate['flux'][ifix.nonzero()])**2
    #
    # Find the list of unique star types
    #
    isort = np.argsort(slist['class'])
    classlist = slist['class'][isort[uniq(slist['class'][isort])]]
    #
    # Loop over each star type
    #
    fullflux = None
    namearr = list()
    colorvec = ['k','r','g','b','m','c']
    smallfont = FontProperties(size='xx-small');
    for c in classlist:
        #
        # Find the subclasses for this stellar type
        #
        print("Finding eigenspectra for Stellar class {0}".format(c))
        indx = (slist['class'] == c).nonzero()[0]
        nindx = indx.size
        thesesubclass = slist['subclass'][indx]
        isort = np.argsort(thesesubclass)
        subclasslist = thesesubclass[isort[uniq(thesesubclass[isort])]]
        nsubclass = subclasslist.size
        #
        # Solve for 2 eigencomponents if we have specified subclasses for
        # this stellar type
        #
        if nsubclass == 1:
            nkeep = 1
        else:
            nkeep = 2
        newloglam = spplate['loglam'][0,:]
        pcaflux = pca_solve(spplate['flux'][indx,:],objinvvar[indx,:],
            spplate['loglam'][indx,:], slist['cz'][indx]/cspeed, wavemin=wavemin,
            wavemax=wavemax, niter=niter, nkeep=nkeep, newloglam=newloglam)
        #
        # Interpolate over bad flux values in the middle of a spectrum,
        # and set fluxes to zero at the blue+red ends of the spectrum
        #
        # minuse = 1 # ?
        minuse = np.floor((nindx+1) / 3.0)
        qbad = pcaflux['usemask'] < minuse
        #
        # Interpolate over all bad pixels
        #
        for j in range(nkeep):
            pcaflux['flux'][j,:] = djs_maskinterp(pcaflux['flux'][j,:],qbad,const=True)
        #
        # Set bad pixels at the very start or end of the spectrum to zero instead
        #
        npix = qbad.size
        igood = (~qbad).nonzero()[0]
        if qbad[0]:
            pcaflux['flux'][:,0:igood[0]-1] = 0
        if qbad[npix-1]:
            pcaflux['flux'][:,igood[::-1][0]+1:npix] = 0
        #
        # Re-normalize the first eigenspectrum to a mean of 1
        #
        # print(pcaflux['flux'].dtype)
        norm = pcaflux['flux'][0,:].mean()
        pcaflux['flux'] /= norm
        pcaflux['acoeff'] *= norm
        # print(pcaflux['flux'].dtype)
        #
        # Now loop through each stellar subclass and reconstruct
        # an eigenspectrum for that subclass
        #
        thesesubclassnum = np.zeros(thesesubclass.size,dtype='i4')
        #
        # Create a figure for this class
        #
        fig = pylab.figure(dpi=100)
        ax = fig.add_subplot(111)
        for isub in range(nsubclass):
            ii = (thesesubclass == subclasslist[isub]).nonzero()[0]
            thesesubclassnum[ii] = isub
            if nkeep == 1:
                thisflux = pcaflux['flux'].reshape(pcaflux['newloglam'].shape)
                # print(thisflux.dtype)
            else:
                aratio = pcaflux['acoeff'][ii,1]/pcaflux['acoeff'][ii,0]
                #
                # np.median(foo) is equivalent to MEDIAN(foo,/EVEN)
                #
                thisratio = np.median(aratio)
                thisflux = pcaflux['flux'][0,:] + thisratio.astype('f') * pcaflux['flux'][1,:]
                # print(thisflux.dtype)
            if fullflux is None:
                fullflux = thisflux
            else:
                fullflux = np.vstack((fullflux,thisflux))
            namearr.append(subclasslist[isub])
            #
            # Plot spectra
            #
            plotflux = thisflux/thisflux.max()
            # print(pcaflux['newloglam'].shape)
            # print(plotflux.shape)
            ax.plot(10.0**pcaflux['newloglam'],plotflux,"{0}-".format(colorvec[isub%len(colorvec)]),linewidth=1)
            if isub == 0:
                ax.set_xlabel(r'Wavelength [$\AA$]')
                ax.set_ylabel('Flux [arbitrary units]')
                ax.set_title('STAR {0}: Eigenspectra Reconstructions'.format(c))
            t = ax.text(10.0**pcaflux['newloglam'][-1],plotflux[-1],subclasslist[isub],
                horizontalalignment='right',verticalalignment='center',
                color=colorvec[isub%len(colorvec)],fontproperties=smallfont)
        fig.savefig(outfile+'.{0}.png'.format(c))
        pylab.close(fig)
        #
        # Plot eigenvalue ratios if nkeep > 1
        #
        if nkeep > 1:
            fig = pylab.figure(dpi=100)
            ax = fig.add_subplot(111)
            allratio = pcaflux['acoeff'][:,1]/pcaflux['acoeff'][:,0]
            isort = thesesubclassnum.argsort()
            p = ax.plot(thesesubclassnum[isort],allratio[isort],marker='None',linestyle='None')
            for k in range(len(indx)):
                t = ax.text(thesesubclassnum[isort[k]],allratio[isort[k]],
                    "%04d-%04d"%(slist['plate'][indx[isort[k]]],slist['fiberid'][indx[isort[k]]]),
                    horizontalalignment='center', verticalalignment='center',
                    color=colorvec[k%len(colorvec)],
                    fontproperties=smallfont)
            ax.set_xlabel('Subclass')
            ax.set_xticks(np.arange(nsubclass))
            ax.set_xticklabels(subclasslist)
            ax.set_ylabel('Eigenvalue Ratio, $a_1/a_0$')
            ax.set_title('STAR {0}: Eigenvalue Ratios'.format(c))
            fig.savefig(outfile+'.{0}.ratio.png'.format(c))
            pylab.close(fig)
    #
    # Save output to FITS file.
    #
    if os.path.exists(outfile+'.fits'):
        os.remove(outfile+'.fits')
    hdu0 = pyfits.PrimaryHDU(fullflux)
    hdu1 = pyfits.new_table(pyfits.ColDefs([
        pyfits.Column(name='plate',format='J',array=slist['plate']),
        pyfits.Column(name='mjd',format='J',array=slist['mjd']),
        pyfits.Column(name='fiber',format='J',array=slist['fiberid']),
        pyfits.Column(name='redshift',format='D',unit='km/s',array=slist['cz'])]))
    hdulist = pyfits.HDUList([hdu0,hdu1])
    hdulist[0].header.update('OBJECT','STAR')
    hdulist[0].header.update('COEFF0',pcaflux['newloglam'][0])
    hdulist[0].header.update('COEFF1',objdloglam)
    hdulist[0].header.update('IDLUTILS','pydlutils','Version of idlutils')
    hdulist[0].header.update('SPEC2D','eigenspectra','Version of idlspec2d')
    hdulist[0].header.update('RUN2D',os.getenv('RUN2D'),'Version of 2d reduction')
    hdulist[0].header.update('RUN1D',os.getenv('RUN1D'),'Version of 1d reduction')
    for i in range(len(namearr)):
        hdulist[0].header.update("NAME%d" % i,namearr[i]+' ')
    hdulist[1].header.update('FILENAME',inputfile)
    hdulist.writeto(outfile+'.fits')
    # plot_eig(outfile+'.fits')
    return



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
import numpy as np
from astropy.utils import lazyproperty
#
class computechi2(object):
    """Solve the linear set of equations Ax=b using SVD.

    The attributes of this class are all implemented as read-only (lazy)properties.

    Parameters
    ----------
    bvec : ndarray
        The b vector in Ax=b. This vector has length N.
    sqivar : ndarray
        The reciprocal of the errors in `b`.  The name comes from the square
        root of the inverse variance, which is what this is.
    amatrix : ndarray
        The matrix A in Ax=b.  The shape of this matrix is (N,M).

    Attributes
    ----------
    chi2 : float
        The chi**2 value of the fit.
    acoeff : ndarray
        The fit parameters, x, in Ax=b.  This vector has length M.
    yfit : ndarray
        The evaluated best-fit at each point.  This vector has length N.
    dof : int
        The degrees of freedom of the fit.  This is the number of values of `bvec`
        that have `sqivar` > 0 minus the number of fit paramaters, which is
        equal to M.
    covar : ndarray
        The covariance matrix.  The shape of this matrix is (M,M).
    var : ndarray
        The variances of the fit.  This is identical to the diagonal of the
        covariance matrix.  This vector has length M.
    """
    #
    #
    #
    def __init__(self,bvec,sqivar,amatrix):
        """Initialize the object and perform initial computations.
        """
        from numpy.linalg import svd
        #
        # Save the inputs
        #
        #self.bvec = bvec
        self.sqivar = sqivar
        self.amatrix = amatrix
        if len(amatrix.shape) > 1:
            self.nstar = amatrix.shape[1]
        else:
            self.nstar = 1
        self.bvec = bvec*sqivar
        self.mmatrix = self.amatrix * np.tile(sqivar,self.nstar).reshape(self.nstar,bvec.size).transpose()
        mm = np.dot(self.mmatrix.T,self.mmatrix)
        self.uu,self.ww,self.vv = svd(mm,full_matrices=False)
        self.mmi = np.dot((self.vv.T / np.tile(self.ww,self.nstar).reshape(self.nstar,self.nstar)),self.uu.T)
        return
    #
    #
    #
    @lazyproperty
    def acoeff(self):
        """Computes the x values in Ax=b.
        """
        return np.dot(self.mmi,np.dot(self.mmatrix.T,self.bvec))
    #
    #
    #
    @lazyproperty
    def chi2(self):
        """Computes the chi**2 value.
        """
        return np.sum((np.dot(self.mmatrix,self.acoeff) - self.bvec)**2)
    #
    #
    #
    @lazyproperty
    def yfit(self):
        """Computes the best fit.
        """
        return np.dot(self.amatrix,self.acoeff)
    #
    #
    #
    @lazyproperty
    def dof(self):
        """Computes the degrees of freedom.
        """
        return (self.sqivar > 0).sum() - self.nstar
    #
    #
    #
    @lazyproperty
    def covar(self):
        """Computes the covariance matrix.
        """
        wwt = self.ww.copy()
        wwt[self.ww>0] = 1.0/self.ww[self.ww>0]
        covar = np.zeros((self.nstar,self.nstar),dtype=self.ww.dtype)
        for i in range(self.nstar):
            for j in range(i+1):
                covar[i,j] = np.sum(wwt * self.vv[:,i] * self.vv[:,j])
                covar[j,i] = covar[i,j]
        return covar
    #
    #
    #
    @lazyproperty
    def var(self):
        """Computes the variances of the fit parameters.
        """
        return np.diag(self.covar)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def mrdfits(filename,hdu=0,**kwargs):
    """Read a FITS file & return the data.

    Only bintables are supported right now.

    Note: this is really a port of hogg_mrdfits, rather than just mrdfits.
    in the future, this may be split into two modules.
    """
    from astropy.io import fits as pyfits
    if 'range' in kwargs and 'rows' in kwargs:
        print "Keywords 'range' and 'rows' are mutually exclusive."
        return None
    try:
        fits = pyfits.open(filename)
    except IOError:
        try:
            fits = pyfits.open(filename+'.gz')
        except IOError:
            return None
    fits_cols = fits[hdu].columns.names
    fits_hdr = fits[hdu].header.ascardlist()
    data = dict()
    if 'columns' in kwargs:
        cols = kwargs['columns']
    else:
        cols = fits_cols
    if 'rows' in kwargs:
        rowdata = fits[hdu].data[kwargs['rows']]
        for col in cols:
            if col not in fits_cols:
                print "Invalid column: %s!" % col
            else:
                data[col] = rowdata[col]
    else:
        if 'range' in kwargs:
            if kwargs['range'] == 'all':
                r = (0,0)
            else:
                if isinstance(kwargs['range'],int):
                    r = (0,kwargs['range'])
                else:
                    r = kwargs['range']
        else:
            r = (0,0)
        for col in cols:
            if col not in fits_cols:
                print "Invalid column: %s!" % col
            else:
                if r[1] > 0:
                    data[col] = fits[hdu].data.field(col)[r[0]:r[1]]
                else:
                    data[col] = fits[hdu].data.field(col)
    fits.close()
    return (data,fits_hdr)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def set_use_caps(x,cm,polygon_use_caps,add=False,tol=1.0e-10,allow_doubles=False,allow_neg_doubles=False):
    """Set the bits in use_caps for a polygon.

    Parameters
    ----------
    x
    cm
    polygon_use_caps
    add
    tol
    allow_doubles
    allow_neg_doubles

    Returns
    -------

    """
    import numpy as np
    from . import is_cap_used
    if add:
        use_caps = long(polygon_use_caps)
    else:
        use_caps = long(0)
    t2 = tol**2
    use_caps |= long(2)**len(cm) - long(1)
    if not allow_doubles:
        #
        # Check for doubles
        #
        for i in range(len(cm)):
            if is_cap_used(use_caps,i):
                for j in range(i+1,len(cm)):
                    if is_cap_used(use_caps,j):
                        if np.sum(x[i]-x[j])**2 < t2:
                            if ((np.absolute(cm[i]-cm[j]) < tol) or
                                ((cm[i] + cm[j]) < tol and not allow_neg_doubles)):
                                #
                                # Don't use
                                #
                                use_caps -= long(2)**j
    return use_caps




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def find_contiguous(x):
    """Find the longest sequence of contiguous non-zero array elements.

    Parameters
    ----------
    x : ndarray
        A 1d array. A dtype of bool is preferred although any dtype where the
        operation ``if x[k]:`` is well-defined should work.

    Returns
    -------
    find_contiguous : list
        A list of indices of the longest contiguous non-zero sequence.

    Examples
    --------
    >>> import numpy as np
    >>> from pydl.pydlutils.math import find_contiguous
    >>> find_contiguous(np.array([0,1,1,1,0,1,1,0,1]))
    [1, 2, 3]
    """
    contig = list()
    for k in range(x.size):
        if x[k]:
            if len(contig) == 0:
                contig.append([k])
            else:
                if k == contig[-1][-1]+1:
                    contig[-1].append(k)
                else:
                    contig.append([k])
    lengths = [len(c) for c in contig]
    longest = contig[lengths.index(max(lengths))]
    return longest




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
def cholesky_band(l,mininf=0.0,verbose=False):
    """Compute Cholesky decomposition of banded matrix.

    Parameters
    ----------
    l : ndarray
        A matrix on which to perform the Cholesky decomposition.
    mininf : float, optional
        Entries in the `l` matrix are considered negative if they are less
        than this value (default 0.0).
    verbose : bool, optional
        If set to ``True``, print some debugging information.

    Returns
    -------
    cholesky_band : tuple
        If problems were detected, the first item will be the index or
        indexes where the problem was detected, and the second item will simply
        be the input matrix.  If no problems were detected, the first item
        will be -1, and the second item will be the Cholesky decomposition.
    """
    import numpy as np
    lower = l.copy()
    bw,nn = lower.shape
    n = nn - bw
    if verbose:
        print(lower[0,0:n])
    negative = lower[0,0:n] <= mininf
    if negative.any() or not np.all(np.isfinite(lower)):
        if verbose:
            print('Bad entries.')
            print(negative.nonzero()[0])
        return (negative.nonzero()[0],l)
    kn = bw - 1
    spot = np.arange(kn,dtype='i4') + 1
    bi = np.arange(kn,dtype='i4')
    for i in range(1,kn):
        bi = np.append(bi, np.arange(kn-i,dtype='i4') + (kn+1)*i)
    for j in range(n):
        lower[0,j] = np.sqrt(lower[0,j])
        lower[spot,j] /= lower[0,j]
        x = lower[spot,j]
        if not np.all(np.isfinite(x)):
            if verbose:
                print('NaN found in cholesky_band.')
            return (j,l)
        hmm = np.outer(x,x)
        here = bi+(j+1)*bw
        lower.T.flat[here] -= hmm.flat[bi]
    return (-1,lower)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
#
__doctest_skip__ = ['read_ds_cooling']
#
def read_ds_cooling(fname,logT=None):
    """Read in Dopita & Sutherland 1993 cooling function.

    Parameters
    ----------
    fname : { 'm-00.cie', 'm-05.cie', 'm+05.cie', 'm-10.cie', 'm-15.cie', 'm-20.cie', 'm-30.cie', 'mzero.cie' }
        Name of the data file to read.
    logT : ndarray, optional
        If provided, values will be interpolated to the provided values.  If
        not provided, the values in the data files will be returned.

    Returns
    -------
    read_ds_cooling : tuple
        A tuple containing `logT` and `loglambda`, respectively.

    Raises
    ------
    URLError
        If the data file could not be retrieved.
    ValueError
        If the input file name is invalid.

    Notes
    -----
    Retrieves data from http://www.mso.anu.edu.au/~ralph/data/cool/ and caches
    the data locally.

    Examples
    --------
    >>> from pydl.pydlutils.cooling import read_ds_cooling
    >>> logT, loglambda = read_ds_cooling('m-15.cie')
    >>> logT[0:5]
    array([ 4.  ,  4.05,  4.1 ,  4.15,  4.2 ])
    >>> loglambda[0:5]
    array([-26.  , -24.66, -23.52, -22.62, -22.11])
    """
    from astropy.utils.data import download_file
    from numpy import interp
    from astropy.io import ascii
    baseurl = 'http://www.mso.anu.edu.au/~ralph/data/cool/'
    if fname not in ('m-00.cie', 'm-05.cie', 'm+05.cie', 'm-10.cie', 'm-15.cie', 'm-20.cie', 'm-30.cie', 'mzero.cie'):
        raise ValueError('Invalid value for data file: {0}'.format(fname))
    filename = download_file(baseurl+fname,cache=True,show_progress=False)
    with open(filename) as coolingfile:
        coolingfile = coolingfile.read()
    data = ascii.read(coolingfile.split('\n')[2:],delimiter='\t')
    if logT is None:
        return (data['log(T)'].data, data['log(lambda net)'].data)
    else:
        loglambda = interp(logT, data['log(T)'].data, data['log(lambda net)'].data)
        return (logT,loglambda)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_yanny():
    """Used to test the yanny class.
    """
    import os
    from .. import yanny
    from numpy import dtype
    #from pydl.pydlutils.yanny import yanny
    #
    # Describe what should be in the object
    #
    pair_dict = {'mjd':'54579','alpha':'beta gamma delta','semicolon':'This pair contains a semicolon;'}
    struct_dict = {
        'MYSTRUCT':{
            'dtype':[('mag', '<f4', (5,)), ('b', 'S33', (5,)), ('foo', 'S25'), ('c', '<f8'), ('flags', '<i4', (2,)), ('new_flag', 'S5')],
            'size':7,
            'columns':{'mag':'float[5]','b':'char[5][]','foo':'char[25]','c':'double','flags':'int[2]','new_flag':'BOOLEAN'},
            },
        'OLD':{
            'dtype':[('foo', '<f4', (3,)), ('bar', 'S10')],
            'size':2,
            'columns':{'foo':'float[3]','bar':'char[10]'},
            },
        'STATUS_UPDATE':{
            'dtype':[('state', 'S10'), ('timestamp', 'S19')],
            'size':11,
            'columns':{'state':'STATUS','timestamp':'char[]'},
            },
        }
    symbols = [
"""typedef struct {
    float mag[5];
    char b[5][];
    char foo[25];
    double c;
    int flags[2];
    BOOLEAN new_flag;
} MYSTRUCT;""",
"""typedef struct {
    float foo<3>; # This is archaic array notation, strongly deprecated,
    char bar<10>; # but still technically supported.
} OLD;""",
"""typedef struct {
    STATUS state;
    char timestamp[]; #UTC timestamp in format 2008-06-21T00:27:33
} STATUS_UPDATE;""",
        ]
    enum = [
"""typedef enum {
    FALSE,
    TRUE
} BOOLEAN;""",
"""typedef enum {
    FAILURE,
    INCOMPLETE,
    SUCCESS
} STATUS;""",
        ]
    #
    # Open the object
    #
    par = yanny(os.path.join(os.path.dirname(__file__),'t','test.par'),np=True)
    #
    # Test the pairs
    #
    assert set(par.pairs()) == set(pair_dict.keys())
    #
    # Test the pair values
    #
    for p in par.pairs():
        assert par[p] == pair_dict[p]
    #
    # Test the structure of the object
    #
    assert set(par.keys()) == set(pair_dict.keys()+struct_dict.keys()+['symbols'])
    assert set(par['symbols'].keys()) == set(struct_dict.keys()+['struct','enum'])
    assert set(par['symbols']['struct']) == set(symbols)
    assert set(par['symbols']['enum']) == set(enum)
    assert set(par.tables()) == set(struct_dict.keys())
    for t in par.tables():
        assert par.dtype(t) == dtype(struct_dict[t]['dtype'])
        assert par.size(t) == struct_dict[t]['size']
        assert set(par.columns(t)) == set(struct_dict[t]['columns'].keys())
        for c in par.columns(t):
            assert par.type(t,c) == struct_dict[t]['columns'][c]
            #print(par[t][c])
    assert par.isenum('MYSTRUCT','new_flag')
    assert par._enum_cache['BOOLEAN'] == ['FALSE','TRUE']
    assert par._enum_cache['STATUS'] == ['FAILURE','INCOMPLETE','SUCCESS']
    #par.write() # This should fail, since test.par already exists.
    #datatable = {'status_update': {'state':['SUCCESS', 'SUCCESS'],
    #    'timestamp':['2008-06-22 01:27:33','2008-06-22 01:27:36']},
    #    'new_keyword':'new_value'}
    #par.filename = os.path.join(os.path.dirname(__file__),'t','test_append.par')
    #par.append(datatable) # This should also fail, because test_append.par does not exist
    return
#
# Testing purposes
#
if __name__ == '__main__':
    test_yanny()




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def cholesky_solve(a,bb):
    """Solve the equation Ax=b where A is a Cholesky-banded matrix.

    Parameters
    ----------
    a
    bb

    Returns
    -------
    cholesky_solve :
    """
    import numpy as np
    b = bb.copy()
    bw = a.shape[0]
    n = b.shape[0] - bw
    kn = bw -1
    spot = np.arange(kn,dtype='i4') + 1
    for j in range(n):
        b[j] /= a[0,j]
        b[j+spot] -= b[j]*a[spot,j]
    spot = kn - np.arange(kn,dtype='i4')
    for j in range(n-1,-1,-1):
        b[j] = (b[j] - np.sum(a[spot,j] * b[j+spot]))/a[0,j]
    return (-1,b)




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def gcirc(ra1,dec1,ra2,dec2,units=2):
    """Computes rigorous great circle arc distances.

    Parameters
    ----------
    ra1, dec1, ra2, dec2 : float or array_like
        RA and Dec of two points.
    units : { 0, 1, 2 }, optional
        * units = 0: everything is already in radians
        * units = 1: RA in hours, dec in degrees, distance in arcsec.
        * units = 2: RA, dec in degrees, distance in arcsec (default)

    Returns
    -------
    gcirc : float or array_like
        The angular distance.  Units of the value returned depend on the
        input value of `units`.

    Notes
    -----
    The formula below is the one best suited to handling small angular
    separations.  See:
    http://en.wikipedia.org/wiki/Great-circle_distance
    """
    from numpy import arcsin, cos, deg2rad, rad2deg, sin, sqrt
    if units == 0:
        rarad1 = ra1
        dcrad1 = dec1
        rarad2 = ra2
        dcrad2 = dec2
    elif units == 1:
        rarad1 = deg2rad(15.0*ra1)
        dcrad1 = deg2rad(dec1)
        rarad2 = deg2rad(15.0*ra2)
        dcrad2 = deg2rad(dec2)
    elif units == 2:
        rarad1 = deg2rad(ra1)
        dcrad1 = deg2rad(dec1)
        rarad2 = deg2rad(ra2)
        dcrad2 = deg2rad(dec2)
    else:
        raise ValueError('units must be 0, 1 or 2!')
    deldec2 = (dcrad2-dcrad1)/2.0
    delra2 =  (rarad2-rarad1)/2.0
    sindis = sqrt( sin(deldec2)*sin(deldec2) +
        cos(dcrad1)*cos(dcrad2)*sin(delra2)*sin(delra2) )
    dis = 2.0*arcsin(sindis)
    if units == 0:
        return dis
    else:
        return rad2deg(dis)*3600.0




#
#
#
def djs_readcol(name,**kwargs):
    """Read a free-format ASCII file into pylab arrays.

    Lines that do not match the specified format are ignored
    (such as comments). Columns may be separated by commas or whitespace.
    Format codes:
    A = string
    B = byte
    D = double precision (float64)
    F = single precision (float32)
    I = integer (int16) (for compatibility with IDL)
    L = long (int32) (for compatibility with IDL)
    K = long64 (int64) (new!)
    X = skip
    """
    import re
    import numpy as np
    #
    # Number of lines
    #
    try:
        f = open(name,'r')
    except IOError:
        return None
    lines = f.readlines()
    f.close()
    nlines = len(lines)
    if 'silent' in kwargs:
        silent = True
    else:
        silent = False
    if 'debug' in kwargs:
        debug = True
    else:
        debug = False
    if debug:
        print "%s contains %d lines." % (name,nlines)
    if 'skip' in kwargs:
        skip = kwargs['skip']
    else:
        skip = 0
    nlines -= skip
    if 'numline' in kwargs:
        nlines = min(kwargs['numline'],nlines)
    #
    # Get the number of columns from the first non-skipped line
    #
    k = skip
    while lines[k][0] == '#':
        k += 1
    whitespace = re.compile(r'\s+')
    baseline = lines[k].strip().replace(',',' ')
    basecols = whitespace.split(baseline)
    ncol = len(basecols)
    if 'format' in kwargs:
        if re.match(r'^\(?[ABDFILX, ]+\)?$',kwargs['format'],re.IGNORECASE) is None:
            print "Invalid format string!"
            return None
        format = kwargs['format'].replace(' ','').upper().lstrip('(').rstrip(')').split(',')
        saveformat = [f for f in format if f != 'X']
        if len(format) < ncol:
            if not silent:
                print 'Format string has fewer columns than the file.'
            ncol = len(format)
    else:
        #
        # Assume all floating point format
        #
        format = list('F'*ncol)
        saveformat = format
    if debug:
        print ','.join(format)
    nread = 0
    goodlist = list()
    for l in lines[skip:nlines]:
        nread += 1
        if debug:
            print l
        if len(l) < ncol or l[0] == '#':
            if not silent:
                print 'Skipping line %d' % (skip+nread+1,)
            continue
        #
        # Split the line
        #
        cols = whitespace.split(l.strip().replace(',',' '))
        savecols = [cols[k] for k in range(ncol) if format[k] != 'X']
        savelist = list()
        if len(savecols) == len(saveformat):
            for k in range(len(saveformat)):
                if saveformat[k] == 'A':
                    #
                    # Save strings as is.
                    #
                    saved = savecols[k]
                elif saveformat[k] == 'B' or saveformat[k] == 'I' or saveformat[k] == 'L':
                    try:
                        saved = int(savecols[k])
                    except ValueError:
                        #
                        # Error, bad format, skip this line
                        #
                        if not silent:
                            print 'Skipping line %d' % skip+nread+1
                        continue
                elif saveformat[k] == 'F' or saveformat[k] == 'D':
                    try:
                        saved = float(savecols[k])
                    except ValueError:
                        #
                        # Error, bad format, skip this line
                        #
                        if not silent:
                            print 'Skipping line %d' % skip+nread+1
                        continue
                else:
                    print "Whoops, bad format! How did that happen?"
                    continue
                savelist.append(saved)
            if len(savelist) != len(saveformat):
                if not silent:
                    print "Skipping line %d" % skip+nread+1
        else:
            #
            # Error, not enough columns
            #
            if not silent:
                print "Skipping line %d" % skip+nread+1
            continue
        goodlist.append(savelist)
    if len(goodlist) == 0:
        raise IOError('No valid lines found for specified format')
    if not silent:
        print "%d valid lines read." % len(goodlist)
    #
    # Zip the good list
    #
    goodcols = zip(*goodlist)
    #
    # Convert the columns to pylab arrays
    #
    dtypes = { 'A':'S','B':'b','I':'i2','L':'i4','K':'i8','F':'f','D':'d' }
    converted = [np.array(goodcols[k],dtype=dtypes[saveformat[k]])
        for k in range(len(saveformat))]
    return tuple(converted)




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
#
from astropy.tests.helper import remote_data
#
@remote_data
def test_sdss_astrombad():
    from .. import sdss_astrombad
    assert sdss_astrombad(77,1,20) == False
    assert sdss_astrombad(77,3,35) == True
    assert sdss_astrombad(77,6,77) == False



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def sdss_calibv():
    """Return calibration for velocities from pix/frame to deg/day.

    Parameters
    ----------
    None

    Returns
    -------
    sdss_calibv : float
        The conversion from pixels per frame to degrees per day

    Notes
    -----
    Assumes frame time difference of 71.72 seconds and pixel scale of 0.396 arcsec, both fixed.
    Also note that observations of the same part of sky from adjacent bands are separated by *two* frame numbers,
    so we multiply by a factor two.
    """
    pixscale = 0.396  # arcsec
    ftime = 71.72    # seconds
    pixframe2degday= 2.0*pixscale/(3600.0) * (3600.0)*24.0/ftime
    return pixframe2degday



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
#
def test_uniq():
    from ..uniq import uniq
    from numpy import array, sort
    items = array([ 1, 2, 3, 1, 5, 6, 1, 7, 3, 2, 5, 9, 11, 1 ])
    u = uniq(sort(items))
    equality = u == array([ 3,  5,  7,  9, 10, 11, 12, 13])
    assert equality.all()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import print_function

from astropy.tests.helper import pytest, remote_data

# this contains imports plugins that configure py.test for astropy tests.
# by importing them here in conftest.py they are discoverable by py.test
# no matter how it is invoked within the astropy tree.

from astropy.tests.pytest_plugins import *



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def sdss_name(ftype, run, camcol, field, rerun='', thisfilter='r',no_path=False):
    """Return the name of an SDSS data file including path.

    Parameters
    ----------
    ftype : str
        The general type of the file, for example ``'reObj'``
    run : int
        The run number.
    camcol : int
        The camcol number.
    field : int
        The field number
    rerun : str, optional
        If necessary, set the rerun name using this argument.
    thisfilter : int or str, optional
        If necessary, set the filter using this argument.
    no_path : bool, optional
        Normally, sdss_name returns the full path.  If `no_path` is ``True``, only
        the basename of the file is returned.

    Returns
    -------
    sdss_name : str
        The full file name, normally including the full path.

    Raises
    ------
    KeyError
        If the file type is unknown.

    """
    from os import getenv
    from os.path import join
    from . import _name_formats, sdss_path
    cname = ('u','g','r','i','z')
    camrow = (3,5,1,2,4)
    if ftype == 'reObj':
        if getenv('PHOTO_RESOLVE') is None:
            myftype = 'reObjRun'
        else:
            myftype = 'reObjGlobal'
    else:
        myftype = ftype
    thisfilter = filtername(thisfilter)
    indict = {
        'ftype':ftype, 'run':run, 'camcol':camcol, 'field':field,
        'filter':thisfilter, 'rerun':rerun
        }
    try:
        fullname = _name_formats[ftype].format(**indict)
    except KeyError:
        raise KeyError("Unknown FTYPE = {0}".format(ftype))
    if not no_path:
        datadir = sdss_path(ftype, run, camcol, rerun)
        fullname = join(datadir, fullname)
    return fullname



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def set_maskbits(idlutils_version='v5_5_8'):
    """Populate the maskbits cache.

    Parameters
    ----------
    idlutils_version : str, optional
        Fetch the sdssMaskbits.par file corresponding to this idlutils version.

    Returns
    -------
    set_maskbits : dict
        A dictionary of bitmasks suitable for caching.

    Raises
    ------
    URLError
        If the data file could not be retrieved.
    """
    from ..yanny import yanny
    #from urllib2 import urlopen
    from astropy.utils.data import download_file
    if idlutils_version == 'trunk' or idlutils_version.startswith('branches/'):
        iversion = idlutils_version
    else:
        iversion = 'tags/'+idlutils_version
    baseurl = 'http://www.sdss3.org/svn/repo/idlutils/{0}/data/sdss/sdssMaskbits.par'.format(iversion)
    filename = download_file(baseurl,cache=True,show_progress=False)
    #par = urlopen(baseurl)
    maskfile = yanny(filename)
    #par.close()
    #
    # Parse the file & cache the results in maskbits
    #
    maskbits = dict()
    for k in range(maskfile.size('MASKBITS')):
        if maskfile['MASKBITS']['flag'][k] in maskbits:
            maskbits[maskfile['MASKBITS']['flag'][k]][maskfile['MASKBITS']['label'][k]] = maskfile['MASKBITS']['bit'][k]
        else:
            maskbits[maskfile['MASKBITS']['flag'][k]] = {maskfile['MASKBITS']['label'][k]:maskfile['MASKBITS']['bit'][k]}
    if 'MASKALIAS' in maskfile:
        for k in range(maskfile.size('MASKALIAS')):
            maskbits[maskfile['MASKALIAS']['alias'][k]] = maskbits[maskfile['MASKALIAS']['flag'][k]].copy()
    return maskbits





# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def sdssflux2ab(flux,magnitude=False,ivar=False):
    """Convert the SDSS calibrated fluxes (magnitudes) into AB fluxes (magnitudes)

    Parameters
    ----------
    flux : ndarray
        Array of calibrated fluxes or SDSS magnitudes with 5 columns,
        corresponding to the 5 filters u,g,r,i,z
    magnitude : bool, optional
        If set to ``True``, then assume `flux` are SDSS magnitudes instead of linear
        flux units
    ivar : ndarray, optional
        If set, the input fluxes are actually inverse variances.

    Returns
    -------
    sdssflux2ab : ndarray
        Array of fluxes or magnitudes on the AB system.

    Notes
    -----
    Uses the conversions posted by D.Hogg (sdss-calib/845)::

     u(AB,2.5m) = u(2.5m) - 0.042
     g(AB,2.5m) = g(2.5m) + 0.036
     r(AB,2.5m) = r(2.5m) + 0.015
     i(AB,2.5m) = i(2.5m) + 0.013
     z(AB,2.5m) = z(2.5m) - 0.002
    """
    import numpy as np
    #
    # Correction vector, adjust this as necessary
    #
    correction = np.array([-0.042, 0.036, 0.015, 0.013, -0.002])
    abflux = flux.copy()
    if magnitude:
        for i in range(5):
            abflux[:,i] += correction[i]
    else:
        factor = np.exp(-correction/2.5 * np.log(10))
        if ivar:
            factor = 1.0/factor**2
        for i in range(5):
            abflux[:,i] *= factor[i]
    return abflux



#
#
#
def radec_to_munu(ra, dec, **kwargs):
    """Convert from equatorial coordinates to SDSS great circle coordinates.
    """
    import numpy as np
    from pydlutils.coord import stripe_to_eta
    from pydlutils.goddard.misc import cirrange
    if 'stripe' in kwargs:
        node = 95.0
        incl = stripe_to_incl(kwargs['stripe'])
    elif 'node' in kwargs and 'incl' in kwargs:
        node = kwargs['node']
        incl = kwargs['incl']
    else:
        raise ValueError('Must specify either STRIPE or NODE,INCL!')
    if ra.size != dec.size:
        raise ValueError('Number of elements in RA and DEC must agree!')
    sinra = np.sin(np.deg2rad(ra-node))
    cosra = np.cos(np.deg2rad(ra-node))
    sindec = np.sin(np.deg2rad(dec))
    cosdec = np.cos(np.deg2rad(dec))
    sini = np.sin(np.deg2rad(incl))
    cosi = np.cos(np.deg2rad(incl))
    x1 = cosdec * cosra
    y1 = cosdec * sinra
    z1 = sindec
    x2 = x1
    y2 = y1 * cosi + z1 * sini
    z2 = -y1 * sini + z1 * cosi
    mu = cirrange(r2d * np.arctan2(y2,x2) + node)
    nu = r2d * np.arcsin(z2)
    if 'phi' in kwargs:
        sinnu = np.sin(np.deg2rad(nu))
        cosnu = np.cos(np.deg2rad(nu))
        sinmu = np.sin(np.deg2rad(mu-node))
        cosmu = np.cos(np.deg2rad(mu-node))
        phi = np.rad2deg(np.arctan2(cosmu * sini,
            (-sinmu * sinnu * sini + cosnu * cosi)*cosnu))
        return (ra,dec,phi)
    else:
        return (ra,dec)




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_struct_print():
    import numpy as np
    from .. import struct_print
    n = 20
    slist = np.zeros(n,dtype=[('PLATE','i4'),('MJD','i4'),('FIBERID','i4'),('RA','f8'),('DEC','f8'),('MATCHRAD','f4'),('RERUN','S8')])
    slist['RERUN'][0:int(n/2)] = 'v1_2_3'
    slist['RERUN'][int(n/2):n] = 'v1_3_34'
    slist['PLATE'] = np.random.random_integers(10,10000,(n,))
    slist['MJD'] = np.random.random_integers(51000,56000,(n,))
    slist['FIBERID'] = np.random.random_integers(1,1000,(n,))
    slist['RA'] = 360.0*np.random.random((n,))
    slist['DEC'] = 90.0 - np.rad2deg(np.arccos(2.0*np.random.random((n,)) - 1.0))
    slist['MATCHRAD'] = np.random.random((n,))
    #print slist
    #lines,css = struct_print(slist,debug=True)
    #print lines
    #print css
    #lines,css = struct_print(slist,debug=True,html=True)
    #print lines
    #print css
    assert slist.size == n



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def latest_mjd(plate,**kwargs):
    """Find the most recent MJD associated with a plate

    Parameters
    ----------
    plate : int or ndarray
        The plate(s) to examine.

    Returns
    -------
    latest_mjd : ndarray
        An array of MJD values for each plate
    """
    import glob
    import re
    import numpy as np
    from . import spec_path
    if isinstance(plate,int) or isinstance(plate,long) or plate.shape == ():
        platevec = np.array([plate],dtype='i4')
    else:
        platevec = plate
    mjd = np.zeros(len(platevec),dtype='i4')
    mjdre = re.compile(r'spPlate-[0-9]{4}-([0-9]{5}).fits')
    unique_plates = np.unique(platevec)
    paths = spec_path(unique_plates,**kwargs)
    for p,q in zip(paths,unique_plates):
        plateglob = "{0}/spPlate-{1:04d}-*.fits".format(p,q)
        bigmjd = 0
        for f in glob.glob(plateglob):
            thismjd = int(mjdre.search(f).groups()[0])
            if thismjd > bigmjd:
                bigmjd = thismjd
        mjd[platevec == q] = bigmjd
    return mjd



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage contains the Goddard utilities.
"""
#
# Define this early on so that submodules can use it
#
#from .. import PydlException
#class GoddardException(PydlException):
#    pass
#class GoddardException(Exception):
#    pass
#
# Import subpackages
#
from . import astro
#from . import fits
from . import math
from . import misc
#
# Set __all__
#
__all__ = ['astro', 'math', 'misc']



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the spec2d directory of idlspec2d.
"""
from .aesthetics import aesthetics
from .combine1fiber import combine1fiber



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage implements functions from the photoop package.
"""
#
# Define this early on so that submodules can use it
#
#from .. import PydlException
#class Pydlspec2dException(PydlException):
#    pass
class Pydlspec2dException(Exception):
    pass
#
# Import subpackages
#
from . import spec1d
from . import spec2d
#
# Set __all__
#
__all__ = ['Pydlspec2dException', 'spec1d', 'spec2d']



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the astro directory in the Goddard utilities.
"""
from .gcirc import gcirc
from .get_juldate import get_juldate



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the image directory of idlutils.
"""
from .djs_maskinterp import djs_maskinterp



from .angles_to_xyz import angles_to_xyz
from .current_mjd import current_mjd
from .munu_to_radec import munu_to_radec
from .radec_to_munu import radec_to_munu
from .stripe_to_eta import stripe_to_eta
from .stripe_to_incl import stripe_to_incl
from .xyz_to_angles import xyz_to_angles



#
# -*- coding: utf-8 -*-
from __future__ import print_function
#
#
def pca_qso(**kwargs):
    """Wrapper on pca_solve to handle QSO eigenspectra.
    """
    import os
    import os.path
    import pickle
    import pylab
    from astropy.io import fits as pyfits
    import numpy as np
    from matplotlib.font_manager import fontManager, FontProperties
    from ...pydlutils.goddard.astro import get_juldate
    from ...pydlutils.image import djs_maskinterp
    from ...pydlutils.math import djs_median, computechi2
    from ...pydlutils.misc import djs_readcol
    from . import pca_solve, plot_eig, readspec, skymask, wavevector
    if 'inputfile' in kwargs:
        inputfile = kwargs['inputfile']
    else:
        inputfile = os.path.join(os.getenv('IDLSPEC2D_DIR'),
            'templates','eigeninput_qso.dat')
    if 'wavemin' in kwargs:
        wavemin = kwargs['wavemin']
    else:
        wavemin = 460.0
    if 'wavemax' in kwargs:
        wavemax = kwargs['wavemax']
    else:
        wavemax = 10000.0
    snmax = 100.0
    if 'niter' in kwargs:
        niter = kwargs['niter']
    else:
        niter = 200
    nkeep = 4
    minuse = 3
    #
    # Name the output files.
    #
    jd = get_juldate()
    outfile = "spEigenQSO-{0:d}".format(int(jd - 2400000.5))
    #
    # Read the input spectra
    #
    plate,mjd,fiber,zfit = djs_readcol(inputfile,format='(L,L,L,D)')
    spplate = readspec(plate,fiber,mjd=mjd,**kwargs)
    #
    # Insist that all of the requested spectra exist.
    #
    missing = spplate['plugmap']['FIBERID'] == 0
    if missing.any():
        imissing = missing.nonzero()[0]
        for k in imissing:
            print("Missing plate={0:d} mjd={1:d} fiber={2:d}".format(plate[k],mjd[k],fiber[k]))
        raise ValueError("{0:d} missing object(s).".format(missing.sum()))
    #
    # Do not fit where the spectrum may be dominated by sky-sub residuals.
    #
    objinvvar = skymask(spplate['invvar'],spplate['andmask'],spplate['ormask'])
    ifix = spplate['flux']**2 * objinvvar > snmax**2
    if ifix.any():
        objinvvar[ifix.nonzero()] = (snmax/spplate['flux'][ifix.nonzero()])**2
    #
    # Set the new wavelength mapping here.  If the binsz keyword is not set,
    # then bin size is determined from the first spectrum returned by readspec.
    # This is fine in the case where all spectra have the same bin size
    # (though their starting wavelengths may differ).  However, this may not
    # be a safe assumption in the future.
    #
    if 'binsz' in kwargs:
        objdloglam = kwargs['binsz']
    else:
        objdloglam = spplate['loglam'][0,1] - spplate['loglam'][0,0]
    newloglam = wavevector(np.log10(wavemin),np.log10(wavemax),
        binsz=objdloglam)
    #
    # Do PCA solution.
    #
    if 'qso' in kwargs:
        #
        # Solve for one component at a time, like the old pca_qso.pro
        #
        objflux = spplate['flux'].copy()
        objloglam = spplate['loglam'].copy()
        nobj, npix = spplate['flux'].shape
        for ikeep in range(nkeep):
            print("Solving for eigencomponent #{0:d} of {1:d}".format(ikeep+1,nkeep))
            pcaflux1 = pca_solve(objflux,objinvvar,objloglam,zfit,
                niter=niter,nkeep=1,newloglam=newloglam,aesthetics='mean')
            if ikeep == 0:
                #
                # Create new pcaflux dict
                #
                saveflux = pcaflux1['newflux']
                pcaflux = dict()
                for k in pcaflux1:
                    pcaflux[k] = pcaflux1[k].copy()
            else:
                #
                # Add to existing dict
                #
                # for k in pcaflux1:
                #     pcaflux[k] = np.vstack((pcaflux[k],pcaflux1[k]))
                pcaflux['flux'] = np.vstack((pcaflux['flux'],pcaflux1['flux']))
                pcaflux['eigenval'] = np.concatenate((pcaflux['eigenval'],pcaflux1['eigenval']))
            #
            # Re-solve for the coefficients using all PCA components so far
            #
            acoeff = np.zeros((nobj,ikeep+1),dtype=pcaflux1['acoeff'].dtype)
            for iobj in range(nobj):
                out = computechi2(saveflux[iobj,:],np.sqrt(pcaflux1['newivar'][iobj,:]),
                    pcaflux['flux'].T)
                acoeff[iobj,:] = out['acoeff']
            #
            # Prevent re-binning of spectra on subsequent calls to pca_solve()
            #
            objloglam = None
            if ikeep == 0:
                objflux = saveflux - np.outer(acoeff,pcaflux['flux'])
            else:
                objflux = saveflux - np.dot(acoeff,pcaflux['flux'])
            # objflux = saveflux - np.outer(acoeff,pcaflux['flux'])
            objinvvar = pcaflux1['newivar']
            pcaflux['acoeff'] = acoeff
    else:
        #
        # Do a normal simultaneous PCA solution
        #
        pcaflux = pca_solve(spplate['flux'],objinvvar,spplate['loglam'],zfit,
            niter=niter,nkeep=nkeep,newloglam=newloglam,aesthetics='mean')
    #
    # Fill in bad data with a running median of the good data.
    #
    qgood = pcaflux['usemask'] >= minuse
    medflux = np.zeros(pcaflux['flux'].shape,dtype=pcaflux['flux'].dtype)
    if not qgood.all():
        for i in range(nkeep):
            medflux[i,qgood] = djs_median(pcaflux['flux'][i,qgood],
                width=51,boundary='nearest')
            medflux[i,:] = djs_maskinterp(medflux[i,:],~qgood,const=True)
        pcaflux['flux'][:,~qgood] = medflux[:,~qgood]
    #
    # Dump input fluxes to a file for debugging purposes.
    #
    if 'dump' in kwargs:
        f = open(kwargs['dump'],'w')
        pickle.dump(pcaflux, f)
        f.close()
    #
    # Make plots
    #
    colorvec = ['k','r','g','b','m','c']
    smallfont = FontProperties(size='xx-small');
    nspectra = pcaflux['newflux'].shape[0]
    if 'flux' in kwargs:
        nfluxes = 30
        separation = 5.0
        nplots = nspectra/nfluxes
        if nspectra % nfluxes > 0:
            nplots += 1
        for k in range(nplots):
            istart = k*nfluxes
            iend = min(istart+nfluxes,nspectra) - 1
            fig = pylab.figure(dpi=100)
            ax = fig.add_subplot(111)
            for l in range(istart,iend+1):
                p = ax.plot(10.0**pcaflux['newloglam'],pcaflux['newflux'][l,:]+separation*(l%nfluxes),
                    '%s-'%colorvec[l%len(colorvec)],linewidth=1)
            ax.set_xlabel(r'Wavelength [$\AA$]')
            ax.set_ylabel(r'Flux [$\mathsf{10^{-17} erg\, cm^{-2} s^{-1} \AA^{-1}}$] + Constant')
            ax.set_title('QSOs: Input Spectra %4d-%4d' % (istart+1,iend+1))
            ax.set_ylim(pcaflux['newflux'][istart,:].min(),pcaflux['newflux'][iend-1,:].max()+separation*(nfluxes-1))
            fig.savefig('%s.flux.%04d-%04d.png'%(outfile,istart+1,iend+1))
            pylab.close(fig)
    fig = pylab.figure(dpi=100)
    ax = fig.add_subplot(111)
    p = ax.plot(10.0**pcaflux['newloglam'],(pcaflux['newivar'] == 0).sum(0)/float(nspectra),'k-')
    ax.set_xlabel(r'Wavelength [$\AA$]')
    ax.set_ylabel('Fraction of spectra with missing data')
    ax.set_title('Missing Data')
    fig.savefig(outfile+'.missing.png')
    pylab.close(fig)
    aratio10 = pcaflux['acoeff'][:,1]/pcaflux['acoeff'][:,0]
    aratio20 = pcaflux['acoeff'][:,2]/pcaflux['acoeff'][:,0]
    aratio30 = pcaflux['acoeff'][:,3]/pcaflux['acoeff'][:,0]
    fig = pylab.figure(dpi=100)
    ax = fig.add_subplot(111)
    p = ax.plot(aratio10,aratio20,marker='None',linestyle='None')
    for k in range(len(aratio10)):
        t = ax.text(aratio10[k],aratio20[k],'%04d-%04d'%(plate[k],fiber[k]),
            horizontalalignment='center', verticalalignment='center',
            color=colorvec[k%len(colorvec)],
            fontproperties=smallfont)
    # ax.set_xlim([aratio10.min(),aratio10.max])
    # ax.set_xlim([aratio20.min(),aratio20.max])
    ax.set_xlabel('Eigenvalue Ratio, $a_1/a_0$')
    ax.set_ylabel('Eigenvalue Ratio, $a_2/a_0$')
    ax.set_title('QSOs: Eigenvalue Ratios')
    fig.savefig(outfile+'.a2_v_a1.png')
    pylab.close(fig)
    fig = pylab.figure(dpi=100)
    ax = fig.add_subplot(111)
    p = ax.plot(aratio20,aratio30,marker='None',linestyle='None')
    for k in range(len(aratio10)):
        t = ax.text(aratio20[k],aratio30[k],'%04d-%04d'%(plate[k],fiber[k]),
            horizontalalignment='center', verticalalignment='center',
            color=colorvec[k%len(colorvec)],
            fontproperties=smallfont)
    # ax.set_xlim([aratio10.min(),aratio10.max])
    # ax.set_xlim([aratio20.min(),aratio20.max])
    ax.set_xlabel('Eigenvalue Ratio, $a_2/a_0$')
    ax.set_ylabel('Eigenvalue Ratio, $a_3/a_0$')
    ax.set_title('QSOs: Eigenvalue Ratios')
    fig.savefig(outfile+'.a3_v_a2.png')
    pylab.close(fig)
    #
    # Save output to FITS file.
    #
    if os.path.exists(outfile+'.fits'):
        os.remove(outfile+'.fits')
    hdu0 = pyfits.PrimaryHDU(pcaflux['flux'])
    hdu1 = pyfits.new_table(pyfits.ColDefs([
        pyfits.Column(name='plate',format='J',array=plate),
        pyfits.Column(name='mjd',format='J',array=mjd),
        pyfits.Column(name='fiber',format='J',array=fiber),
        pyfits.Column(name='redshift',format='D',array=zfit)]))
    hdulist = pyfits.HDUList([hdu0,hdu1])
    hdulist[0].header.update('OBJECT','QSO')
    hdulist[0].header.update('COEFF0',pcaflux['newloglam'][0])
    hdulist[0].header.update('COEFF1',pcaflux['newloglam'][1]-pcaflux['newloglam'][0])
    hdulist[0].header.update('IDLUTILS','pydlutils','Version of idlutils')
    hdulist[0].header.update('SPEC2D','eigenspectra','Version of idlspec2d')
    hdulist[0].header.update('RUN2D',os.getenv('RUN2D'),'Version of 2d reduction')
    hdulist[0].header.update('RUN1D',os.getenv('RUN1D'),'Version of 1d reduction')
    for i in range(len(pcaflux['eigenval'])):
        hdulist[0].header.update("EIGEN%d" % i,pcaflux['eigenval'][i])
    hdulist[1].header.update('FILENAME',inputfile)
    hdulist.writeto(outfile+'.fits')
    plot_eig(outfile+'.fits')
    return



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
def hogg_iau_name(ra,dec,prefix='SDSS',precision=1,debug=False):
    """Properly format astronomical source names to the IAU convention.

    Parameters
    ----------
    ra : float or ndarray
        Right ascencion in decimal degrees
    dec : float or ndarray
        Declination in decimal degrees.
    prefix : str, optional
        Add this prefix to the string, defaults to 'SDSS'.
    precision : int, optional
        Display this many digits of precision on seconds, default 1.
    debug : bool, optional
        If ``True``, print some extra debug information.

    Returns
    -------
    hogg_iau_name : str or list
        The IAU name for the coordinates.

    Examples
    --------
    >>> from pydl.pydlutils.misc import hogg_iau_name
    >>> hogg_iau_name(354.120375,-0.544777778)
    'SDSS J233628.89-003241.2'
    """
    import numpy as np
    #
    # Promote scalar values to arrays.
    #
    if isinstance(ra,float):
        ra = np.array([ra])
    if isinstance(dec,float):
        dec = np.array([dec])
    h = ra/15.0
    rah = np.floor(h)
    ram = np.floor(60.0*(h-rah))
    ras = 60.0*(60.0*(h-rah) - ram)
    ras = np.floor(ras*10.0**(precision+1))/10.0**(precision+1)
    rasformat = "{{2:0{0:d}.{1:d}f}}".format(precision+4, precision+1)
    if debug:
        print(rasformat)
    rah = rah.astype(np.int32)
    ram = ram.astype(np.int32)
    desgn = np.array(list('+'*len(dec)))
    desgn[dec < 0] = '-'
    adec = np.absolute(dec)
    ded = np.floor(adec)
    dem = np.floor(60.0*(adec-ded))
    des = 60.0*(60.0*(adec-ded) - dem)
    des = np.floor(des*10.0**precision)/10.0**precision
    desformat = "{{6:0{0:d}.{1:d}f}}".format(precision+3, precision)
    if precision == 0:
        desformat = "{6:02d}"
        des = des.astype(np.int32)
    if debug:
        print(desformat)
    ded = ded.astype(np.int32)
    dem = dem.astype(np.int32)
    adformat = "{{0:02d}}{{1:02d}}{ras}{{3:s}}{{4:02d}}{{5:02d}}{des}".format(ras=rasformat,des=desformat)
    if debug:
        print(adformat)
        print(rah, ram, ras, desgn, ded, dem, des)
        print(rah.dtype, ram.dtype, ras.dtype, desgn.dtype, ded.dtype, dem.dtype, des.dtype)
    #
    # The easy way doesn't work if numpy version is less than 1.7.0.  Prior to this
    # version, numpy scalars didn't support __format__.  See
    # http://projects.scipy.org/numpy/ticket/1675
    #
    try:
        adstr = [adformat.format(*x) for x in zip(rah, ram, ras, desgn, ded, dem, des)]
    except ValueError:
        adstr = [adformat.format(*x) for x in zip(
            rah.tolist(), ram.tolist(), ras.tolist(), desgn.tolist(),
            ded.tolist(), dem.tolist(), des.tolist())]
    if prefix == '':
        jstr = 'J'
    else:
        jstr = ' J'
    name = ["{0}{1}{2}".format(prefix, jstr, x) for x in adstr]
    if len(ra) == 1:
        return name[0]
    else:
        return name
#
#
#
def main():
    from astropy.utils.compat import argparse
    parser = argparse.ArgumentParser(description='Properly format astronomical source names to the IAU convention.')
    parser.add_argument('-d', '--debug', dest='debug', action='store_true',
        help='Print extra debug information.')
    parser.add_argument('-P', '--precision', dest='precision', action='store',
        metavar='N', default=1, type=int, help='Digits of precision to add to the declination.')
    parser.add_argument('-p', '--prefix', dest='prefix', action='store',
        metavar='STR', default='SDSS', help='Add this prefix to the name.')
    parser.add_argument('ra', metavar='RA', type=float,
        help='Right Ascension.')
    parser.add_argument('dec', metavar='Dec', type=float,
        help='Declination.')
    options = parser.parse_args()
    print(hogg_iau_name(options.ra,options.dec,
        prefix=options.prefix,precision=options.precision,debug=options.debug))
    return



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
def skymask(invvar,andmask,ormask=None,ngrow=2):
    """Mask regions where sky-subtraction errors are expected to dominate.
    """
    from numpy import zeros
    from ...pydlutils.sdss import sdss_flagval
    from ... import smooth
    nrows,npix = invvar.shape
    badmask = zeros(invvar.shape,dtype='i4')
    badskychi = sdss_flagval('SPPIXMASK','BADSKYCHI')
    redmonster = sdss_flagval('SPPIXMASK','REDMONSTER')
    # brightsky = sdss_flagval('SPPIXMASK','BRIGHTSKY')
    if ormask is not None:
        badmask = badmask | ((ormask & badskychi) != 0)
        badmask = badmask | ((ormask & redmonster) != 0)
        # badmask = badmask | ((andmask & brightsky) != 0)
    if ngrow > 0:
        width = 2*ngrow + 1
        for k in range(nrows):
            badmask[k,:] = smooth(badmask[k,:]*width, width, True) > 0
    return invvar * (1 - badmask)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the sdssio directory of photoop.
"""
from .filtername import filtername
from .filternum import filternum
from .sdss_name import sdss_name
from .sdss_path import sdss_path
from .sdssflux2ab import sdssflux2ab
#
# Filename formats used by sdss_name and sdss_path
#
_name_formats = {
    'apObj':"{ftype}-{run:06d}-{filter}{camcol:1d}-{filter:04d}.fit",
    'calibMatch':"{ftype}-{run:06d}-{camcol:1d}.fits",
    'calibPhotom':"{ftype}-{run:06d}-{camcol:1d}.fits",
    'calibPhotomGlobal':"{ftype}-{run:06d}-{camcol:1d}.fits",
    'fakeIdR':"idR-{run:06d}-{filter}{camcol:1d}-{filter:04d}.fit",
    'fpAtlas':"{ftype}-{run:06d}-{camcol:1d}-{filter:04d}.fit",
    'fpBIN':"{ftype}-{run:06d}-{filter}{camcol:1d}-{filter:04d}.fit",
    'fpC':"{ftype}-{run:06d}-{filter}{camcol:1d}-{filter:04d}.fit",
    'fpFieldStat':"{ftype}-{run:06d}-{camcol:1d}-{filter:04d}.fit",
    'fpM':"{ftype}-{run:06d}-{filter}{camcol:1d}-{filter:04d}.fit",
    'fpObjc':"{ftype}-{run:06d}-{camcol:1d}-{filter:04d}.fit",
    'hoggObj':"{ftype}-{run:06d}-{camcol:1d}-{filter:04d}.fits",
    'idFF':"{ftype}-{run:06d}-{filter}{camcol:1d}.fit",
    'idR':"{ftype}-{run:06d}-{filter}{camcol:1d}-{filter:04d}.fit",
    'idRR':"{ftype}-{run:06d}-{filter}{camcol:1d}-{filter:04d}.fit",
    'psBB':"{ftype}-{run:06d}-{filter}{camcol:1d}-{filter:04d}.fit",
    'psFF':"{ftype}-{run:06d}-{filter}{camcol:1d}.fit",
    'psField':"{ftype}-{run:06d}-{camcol:1d}-{filter:04d}.fit",
    'tsField':"{ftype}-{run:06d}-{camcol:1d}-{rerun}-{filter:04d}.fit",
    }
#
#
#
_path_formats = {
    'apObj':"{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'calibMatch':"{redux}/{rerun}/{run:d}/nfcalib",
    'calibPhotom':"{redux}/{rerun}/{run:d}/nfcalib",
    'calibPhotomGlobal':"{calib}/{rerun}/{run:d}/nfcalib",
    'fakeIdR':"{data}/{run:d}/fake_fields/{camcol:1d}",
    'fpAtlas':"{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'fpBIN':"{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'fpC':"{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'fpFieldStat':"{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'fpM':"{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'fpObjc':"{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'hoggObj':"{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'idFF':"{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'idR':"{data}/{run:d}/fields/{camcol:1d}",
    'idRR':"{data}/{run:d}/fields/{camcol:1d}",
    'psBB':"{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'psFF':"{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'psField':"{redux}/{rerun}/{run:d}/objcs/{camcol:1d}",
    'tsField':"{redux}/{rerun}/{run:d}/calibChunks/{camcol:1d}",
    }



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
def wavevector(minfullwave,maxfullwave,**kwargs):
    """Simply return an array of wavelengths.
    """
    import numpy as np
    if 'zeropoint' in kwargs:
        zeropoint = kwargs['zeropoint']
    else:
        zeropoint = 3.5
    if 'binsz' in kwargs:
        binsz = kwargs['binsz']
    else:
        binsz = 1.0e-4
    if 'wavemin' in kwargs:
        wavemin = kwargs['wavemin']
        spotmin = 0
        if 'wavemax' in kwargs:
            wavemax = kwargs['wavemax']
            spotmax = int((wavemax - wavemin)/binsz)
        else:
            spotmax = int((maxfullwave - wavemin)/binsz)
            wavemax = spotmax * binsz + wavemin
    else:
        spotmin = int((minfullwave - zeropoint)/binsz) + 1
        spotmax = int((maxfullwave - zeropoint)/binsz)
        wavemin = spotmin * binsz + zeropoint
        wavemax = spotmax * binsz + zeropoint
    nfinalpix = spotmax - spotmin + 1
    finalwave = np.arange(nfinalpix,dtype='d') *binsz + wavemin
    return finalwave




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def djs_reject(data,model,outmask=None,inmask=None,sigma=None,invvar=None,
    lower=None,upper=None,maxdev=None,
    maxrej=None,groupdim=None,groupsize=None,groupbadpix=False,
    grow=0,sticky=False):
    """Routine to reject points when doing an iterative fit to data.

    Parameters
    ----------
    data : ndarray
        The data
    model : ndarray
        The model, must have the same number of dimensions as `data`.
    outmask : ndarray, optional
        Output mask, generated by a previous call to `djs_reject`.  If not supplied,
        this mask will be initialized to a mask that masks nothing.  Although
        this parameter is technically optional, it will almost always be set.
    inmask : ndarray, optional
        Input mask.  Bad points are marked with a value that evaluates to ``False``.
        Must have the same number of dimensions as `data`.
    sigma : ndarray, optional
        Standard deviation of the data, used to reject points based on the values
        of `upper` and `lower`.
    invvar : ndarray, optional
        Inverse variance of the data, used to reject points based on the values
        of `upper` and `lower`.  If both `sigma` and `invvar` are set, `invvar`
        will be ignored.
    lower : int or float, optional
        If set, reject points with data < model - lower * sigma.
    upper : int or float, optional
        If set, reject points with data > model + upper * sigma.
    maxdev : int or float, optional
        If set, reject points with abs(data-model) > maxdev.  It is permitted to
        set all three of `lower`, `upper` and `maxdev`.
    maxrej : int or ndarray, optional
        Maximum number of points to reject in this iteration.  If `groupsize` or
        `groupdim` are set to arrays, this should be an array as well.
    groupdim
    groupsize
    groupbadpix : bool, optional
        If set to ``True``, consecutive sets of bad pixels are considered groups,
        overriding the values of `groupsize`.
    grow : int, optional
        If set to a non-zero integer, N, the N nearest neighbors of rejected
        pixels will also be rejected.
    sticky : bool, optional
        If set to ``True``, pixels rejected in one iteration remain rejected in
        subsequent iterations, even if the model changes.

    Returns
    -------
    djs_reject : tuple
        A tuple containing a mask where rejected data values are ``False`` and
        a boolean value set to ``True`` if `djs_reject` believes there is no
        further rejection to be done.

    Raises
    ------
    ValueError
        If dimensions of various inputs do not match.
    """
    import numpy as np
    from ..misc import djs_laxisnum
    #
    # Create outmask setting = 1 for good data.
    #
    if outmask is None:
        outmask = np.ones(data.shape,dtype='bool')
    else:
        if data.shape != outmask.shape:
            raise ValueError('Dimensions of data and outmask do not agree.')
    #
    # Check other inputs.
    #
    if model is None:
        if inmask is not None:
            outmask = inmask
        return (outmask,False)
    else:
        if data.shape != model.shape:
            raise ValueError('Dimensions of data and model do not agree.')
    if inmask is not None:
        if data.shape != inmask.shape:
            raise ValueError('Dimensions of data and inmask do not agree.')
    if maxrej is not None:
        if groupdim is not None:
            if len(maxrej) != len(groupdim):
                raise ValueError('maxrej and groupdim must have the same number of elements.')
        else:
            groupdim = []
        if groupsize is not None:
            if len(maxrej) != len(groupsize):
                raise ValueError('maxrej and groupsize must have the same number of elements.')
        else:
            groupsize = len(data)
    if sigma is None and invvar is None:
        if inmask is not None:
            igood = (inmask & outmask).nonzero()[0]
        else:
            igood = outmask.nonzero()[0]
        if len(igood > 1):
            sigma = np.std(data[igood] - model[igood])
        else:
            sigma = 0
    diff = data - model
    #
    # The working array is badness, which is set to zero for good points
    # (or points already rejected), and positive values for bad points.
    # The values determine just how bad a point is, either corresponding
    # to the number of sigma above or below the fit, or to the number
    # of multiples of maxdev away from the fit.
    #
    badness = np.zeros(outmask.shape,dtype=data.dtype)
    #
    # Decide how bad a point is according to lower.
    #
    if lower is not None:
        if sigma is not None:
            qbad = diff < (-lower * sigma)
            badness += ((-diff/(sigma + (sigma == 0))) > 0) * qbad
        else:
            qbad = (diff * np.sqrt(invvar)) < -lower
            badness += ((-diff * np.sqrt(invvar)) > 0) * qbad
    #
    # Decide how bad a point is according to upper.
    #
    if upper is not None:
        if sigma is not None:
            qbad = diff > (upper * sigma)
            badness += ((diff/(sigma + (sigma == 0))) > 0) * qbad
        else:
            qbad = (diff * np.sqrt(invvar)) > upper
            badness += ((diff * np.sqrt(invvar)) > 0) * qbad
    #
    # Decide how bad a point is according to maxdev.
    #
    if maxdev is not None:
        qbad = np.absolute(diff) > maxdev
        badness += np.absolute(diff) / maxdev * qbad
    #
    # Do not consider rejecting points that are already rejected by inmask.
    # Do not consider rejecting points that are already rejected by outmask,
    # if sticky is set.
    #
    if inmask is not None:
        badness *= inmask
    if sticky:
        badness *= outmask
    #
    # Reject a maximum of maxrej (additional) points in all the data, or
    # in each group as specified by groupsize, and optionally along each
    # dimension specified by groupdim.
    #
    if maxrej is not None:
        #
        # Loop over each dimension of groupdim or loop once if not set.
        #
        for iloop in range(max(len(groupdim),1)):
            #
            # Assign an index number in this dimension to each data point.
            #
            if len(groupdim) > 0:
                yndim = len(ydata.shape)
                if groupdim[iloop] > yndim:
                    raise ValueError('groupdim is larger than the number of dimensions for ydata.')
                dimnum = djs_laxisnum(ydata.shape,iaxis=groupdim[iloop]-1)
            else:
                dimnum = 0
            #
            # Loop over each vector specified by groupdim. For example, if
            # this is a 2-D array with groupdim=1, then loop over each
            # column of the data.  If groupdim=2, then loop over each row.
            # If groupdim is not set, then use the whole image.
            #
            for ivec in range(max(dimnum)):
                #
                # At this point it is not possible that dimnum is not set.
                #
                indx = (dimnum == ivec).nonzero()[0]
                #
                # Within this group of points, break it down into groups
                # of points specified by groupsize, if set.
                #
                nin = len(indx)
                if groupbadpix:
                    goodtemp = badness == 0
                    groups_lower = (-1*np.diff(np.insert(goodtemp,0,1)) == 1).nonzero()[0]
                    groups_upper = (np.diff(np.append(goodtemp,1)) == 1).nonzero()[0]
                    ngroups = len(groups_lower)
                else:
                    #
                    # The IDL version of this test makes no sense because
                    # groupsize will always be set.
                    #
                    if 'groupsize' in kwargs:
                        ngroups = nin/groupsize + 1
                        groups_lower = np.arange(ngroups,dtype='i4')*groupsize
                        foo = (np.arange(ngroups,dtype='i4')+1)*groupsize
                        groups_upper = np.where(foo < nin,foo,nin) -1
                    else:
                        ngroups = 1
                        groups_lower = [0,]
                        groups_upper = [nin - 1,]
                for igroup in range(ngroups):
                    i1 = groups_lower[igroup]
                    i2 = groups_upper[igroup]
                    nii = i2 - i1 + 1
                    #
                    # Need the test that i1 != -1 below to prevent a crash
                    # condition, but why is it that we ever get groups
                    # without any points?  Because this is badly-written,
                    # that's why.
                    #
                    if nii > 0 and i1 != -1:
                        jj = indx[i1:i2+1]
                        #
                        # Test if too many points rejected in this group.
                        #
                        if np.sum(badness[jj] != 0) > maxrej[iloop]:
                            isort = badness[jj].argsort()
                            #
                            # Make the following points good again.
                            #
                            badness[jj[isort[0:nii-maxrej[iloop]]]] = 0
                        i1 += groupsize[iloop]
    #
    # Now modify outmask, rejecting points specified by inmask=0, outmask=0
    # if sticky is set, or badness > 0.
    #
    # print badness
    newmask = badness == 0
    # print newmask
    if grow > 0:
        rejects = newmask==0
        if rejects.any():
            irejects = rejects.nonzero()[0]
            for k in range(1,grow):
                newmask[(irejects - k) > 0] = 0
                newmask[(irejects + k) < (data.shape[0]-1)] = 0
    if inmask is not None:
        newmask = newmask & inmask
    if sticky:
        newmask = newmask & outmask
    #
    # Set qdone if the input outmask is identical to the output outmask.
    #
    qdone = np.all(newmask == outmask)
    outmask = newmask
    return (outmask,qdone)




#
#
#
def hogg_mrdfits(filename,hdu=0,**kwargs):
    """Read a FITS file & return the data.

    Only bintables are supported right now.

    Note: pydlutils.goddard.fits.mrdfits is the 'real' port of
    hogg_mrdfits at least for now.  This function provides
    a placeholder for further development.
    """
    from pydlutils.goddard.fits import mrdfits
    return mrdfits(filename,hdu,**kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the math directory of idlutils.
"""
from .computechi2 import computechi2
from .djs_median import djs_median
from .djs_reject import djs_reject
from .find_contiguous import find_contiguous



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def djs_median(array,dimension=None,width=None,boundary='none'):
    """Compute the median of an array.

    Use a filtering box or collapse the image along one dimension.

    Parameters
    ----------
    array : ndarray
        input array
    dimension : int, optional
        Compute the median over this dimension. It is an error to specify both
        `dimension` and `width`.
    width : int, optional
        Width of the median window. It is an error to specify both
        `dimension` and `width`.
    boundary : { 'none', 'reflect', 'nearest', 'wrap' }, optional
        Boundary condition to impose.  'none' means no filtering is done within
        `width`/2 of the boundary.  'reflect' means reflect pixel values around the
        boundary. 'nearest' means use the values of the nearest boundary pixel.
        'wrap' means wrap pixel values around the boundary. 'nearest' and 'wrap'
        are not implemented.

    Returns
    -------
    djs_median : ndarray
        The output.  If neither `dimension` nor `width` are set, this is a scalar
        value, just the output of ``numpy.median()``.  If `width` is set, the
        result has the same shape as the input array.
    """
    import numpy as np
    from scipy.signal import medfilt
    if dimension is None and width is None:
        return np.median(array)
    elif dimension is None:
        if boundary == 'none':
            if width == 1:
                medarray = array
            else:
                medarray = medfilt(array,min(width,array.size))
                #
                # medfilt doesn't handle edges the same way as IDL, so we
                # have to fix them up
                #
                istart = (width-1)/2
                iend = array.size - (width+1)/2
                i = np.arange(array.size)
                w = (i < istart) | (i > iend)
                medarray[w] = array[w]
        else:
            padsize = np.ceil(width/2.0)
            if array.ndim == 1:
                bigarr = np.zeros(array.shape[0]+2*padsize,dtype=array.dtype)
                bigarr[padsize:padsize+array.shape[0]] = array
            elif array.ndim == 2:
                bigarr = np.zeros((array.shape[0]+2*padsize,array.shape[1]+2*padsize),dtype=array.dtype)
                bigarr[padsize:padsize+array.shape[0],padsize:padsize+array.shape[1]] = array
            else:
                raise ValueError('Unsupported number of dimensions with this boundary condition.')
            if array.ndim == 1:
                if width == 1:
                    medarray = array
                else:
                    bigarr[0:padsize] = array[0:padsize][::-1]
                    bigarr[padsize+array.shape[0]:padsize*2+array.shape[0]] = array[array.shape[0]-padsize:array.shape[0]]
                    f = medfilt(bigarr,width)
                    istart = (width-1)/2
                    iend = array.size - (width+1)/2
                    i = np.arange(array.size)
                    w = (i < istart) | (i > iend)
                    f[w] = array[w]
                    medarray = f[padsize:padsize+array.shape[0]]
            else:
                if boundary == 'nearest':
                    raise NotImplementedError('This boundary condition not implemented')
                elif boundary == 'wrap':
                    raise NotImplementedError('This boundary condition not implemented')
                elif boundary == 'reflect':
                    # medarray = array
                    raise NotImplementedError('This boundary condition not implemented')
                else:
                    raise ValueError('Unknown boundary condition.')
    elif width is None:
        raise NotImplementedError('This type of median is not implemented')
    else:
        raise ValueError('Invalid to specify both dimension & width.')
    return medarray



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def filtername(f):
    """Return the name of a filter given its number.

    Parameters
    ----------
    f : int
        The filter number.

    Returns
    -------
    filtername : str
        The corresponding filter name.

    Examples
    --------
    >>> filtername(0)
    'u'
    """
    if isinstance(f,str):
        return f
    fname = ('u','g','r','i','z')
    return fname[f]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
import numpy as np
#
#
#
def model(a,g):
    """Compute the model."""
    return np.dot(a,g)
#
#
#
def resid(a,g,spectra):
    """Compute residuals."""
    return spectra - model(a,g)
#
#
#
def chi(a,g,spectra,invvar):
    """Compute chi, the scaled residual."""
    return resid(a,g,spectra)*np.sqrt(invvar)
#
#
#
def penalty(g,epsilon=None):
    """Compute penalty for non-smoothness."""
    if epsilon is None:
        return 0.0
    return epsilon*np.sum(np.diff(g)**2)
#
#
#
def badness(a,g,spectra,invvar,epsilon=None):
    """Compute chi**2."""
    return np.sum(chi(a,g,spectra,invvar)**2) + penalty(g,epsilon)
#
#
#
def normbase(g):
    """Apply standard component normalization."""
    return np.sqrt((g**2).mean(1))
#
#
#
def astep(spectra,invvar,g):
    """Update for coefficients at fixed component spectra."""
    from numpy.linalg import solve
    N,M = spectra.shape
    K,M = g.shape
    a = np.zeros((N,K),dtype=g.dtype)
    for i in range(N):
        Gi = np.zeros((K,K),dtype=g.dtype)
        for k in range(K):
            for kp in range(k,K):
                Gi[k,kp] = np.sum(g[k,:]*g[kp,:]*invvar[i,:])
                if kp > k:
                    Gi[kp,k] = Gi[k,kp]
        Fi = np.dot(g,spectra[i,:]*invvar[i,:])
        a[i,:] = solve(Gi,Fi)
    return a
#
#
#
def gstep(oldg,spectra,invvar,a,epsilon=None):
    """Update for component spectra at fixed coefficients."""
    from numpy.linalg import solve
    N,M = spectra.shape
    N,K = a.shape
    g = np.zeros((K,M),dtype=a.dtype)
    e = np.zeros(oldg.shape,dtype=oldg.dtype)
    d = np.zeros((K,K,M),dtype=a.dtype)
    if epsilon is not None and epsilon > 0:
        foo = epsilon*np.eye(K,dtype=a.dtype)
        for l in range(M):
            d[:,:,l] = foo
            if l > 0 and l < M-1:
                d[:,:,l] *= 2
        # d[:,:,0] = foo
        # d[:,:,1:M-1] = 2*foo
        # d[:,:,M-1] = foo
        e[:,0] = epsilon*oldg[:,1]
        e[:,1:M-1] = epsilon*(oldg[:,0:M-2] + oldg[:,2:M])
        e[:,M-1] = epsilon*oldg[:,M-2]
    for j in range(M):
        Aj = np.zeros((K,K),dtype=a.dtype)
        for k in range(K):
            for kp in range(k,K):
                Aj[k,kp] = np.sum(a[:,k]*a[:,kp]*invvar[:,j])
                if kp > k:
                    Aj[kp,k] = Aj[k,kp]
        Aj += d[:,:,j]
        Fj = np.dot(a.T,spectra[:,j]*invvar[:,j]) + e[:,j]
        g[:,j] = solve(Aj,Fj)
    return g
#
#
#
def astepnn(a,spectra,invvar,g):
    """Non-negative update for coefficients at fixed component spectra."""
    numerator = np.dot(spectra*invvar,g.T)
    denominator = np.dot(np.dot(a,g)*invvar,g.T)
    return a*(numerator/denominator)
#
#
#
def gstepnn(g,spectra,invvar,a,epsilon=None):
    """Non-negative update for component spectra at fixed coefficients."""
    K,M = g.shape
    numerator = np.dot(a.T,(spectra*invvar))
    if epsilon is not None and epsilon > 0:
        e = np.zeros(g.shape,dtype=g.dtype)
        e[:,0] = epsilon*g[:,1]
        e[:,1:M-1] = epsilon*(g[:,0:M-2] + g[:,2:M])
        e[:,M-1] = epsilon*g[:,M-2]
        numerator += e
    denominator = np.dot(a.T,np.dot(a,g)*invvar)
    if epsilon is not None and epsilon > 0:
        d = epsilon*g.copy()
        d[:,1:M-1] *= 2
        denominator += d
    return g*(numerator/denominator)
#
#
#
def reorder(a,g):
    """Reorder and rotate basis analogous to PCA."""
    from numpy.linalg import eigh
    l,U = eigh(np.dot(a.T,a))
    return (np.dot(a,U),np.dot(U.T,g))
#
#
#
def hmf_solve(spectra,invvar,K=4,nonnegative=False,epsilon=None):
    """Handle the HMF iteration, analogous to pca_solve.

    Parameters
    ----------
    spectra : numpy.array
        The training spectra.
    invvar : numpy.array
        The inverse variance for each pixel in the spectra.
    K : int, optional
        The number of dimensions of the factorization (default 4).
    nonnegative : bool, optional
        Set this to ``True`` to perform nonnegative HMF.
    epsilon : float, optional
        Regularization parameter.  Set to any non-zero float value to turn it on.

    Returns
    -------
    a,g : tuple of numpy.array
        The fitting coefficients and fitted functions, respectively.
    """
    from scipy.cluster.vq import kmeans, whiten
    N,M = spectra.shape
    #
    # Make spectra non-negative
    #
    if nonnegative:
        spectra[spectra<0] = 0
        invvar[spectra<0] = 0
    #
    # Detect very bad columns
    #
    si = spectra*invvar
    if (spectra.sum(0) == 0).any():
        raise ValueError("Columns of zeros detected in spectra!")
    if (invvar.sum(0) == 0).any():
        raise ValueError("Columns of zeros detected in invvar!")
    if (si.sum(0) == 0).any():
        raise ValueError("Columns of zeros detected in spectra*invvar!")
    #
    # Initialize g matrix with kmeans
    #
    whitespectra = whiten(spectra)
    g,foo = kmeans(whitespectra,K)
    g /= np.repeat(normbase(g),M).reshape(g.shape)
    #
    # Initialize a matrix
    #
    a = np.outer(np.sqrt((spectra**2).mean(1)),np.repeat(1.0/K,K))
    if nonnegative:
        for k in range(128):
            a = astepnn(a,spectra,invvar,g)
    #
    # Number of iterations.
    #
    if nonnegative:
        n_iter = 2048
    else:
        n_iter = 16
    #
    # Iterate!
    #
    for m in range(n_iter):
        print(m)
        if nonnegative:
            a = astepnn(a,spectra,invvar,g)
            g = gstepnn(g,spectra,invvar,a,epsilon)
        else:
            a = astep(spectra,invvar,g)
            g = gstep(g,spectra,invvar,a,epsilon)
            a,g = reorder(a,g)
        norm = normbase(g)
        g /= np.repeat(norm,M).reshape(g.shape)
        a = (a.T*np.repeat(norm,N).reshape(K,N)).T
        # print(badness(a,g,spectra,invvar,epsilon))
    return (a,g)



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst

import glob
import os
import sys

import setuptools_bootstrap
from setuptools import setup

#A dirty hack to get around some early import/configurations ambiguities
if sys.version_info[0] >= 3:
    import builtins
else:
    import __builtin__ as builtins
builtins._ASTROPY_SETUP_ = True

import astropy
from astropy.setup_helpers import (register_commands, adjust_compiler,
                                   get_debug_option)
from astropy.version_helpers import get_git_devstr, generate_version_py

# Get some values from the setup.cfg
from distutils import config
conf = config.ConfigParser()
conf.read(['setup.cfg'])
metadata = dict(conf.items('metadata'))

PACKAGENAME = metadata.get('package_name', 'packagename')
DESCRIPTION = metadata.get('description', 'Astropy affiliated package')
AUTHOR = metadata.get('author', '')
AUTHOR_EMAIL = metadata.get('author_email', '')
LICENSE = metadata.get('license', 'unknown')
URL = metadata.get('url', 'http://astropy.org')

# Get the long description from the package's docstring
__import__(PACKAGENAME)
package = sys.modules[PACKAGENAME]
LONG_DESCRIPTION = package.__doc__

# Store the package name in a built-in variable so it's easy
# to get from other parts of the setup infrastructure
builtins._ASTROPY_PACKAGE_NAME_ = PACKAGENAME

# VERSION should be PEP386 compatible (http://www.python.org/dev/peps/pep-0386)
VERSION = '0.2.2'

# Indicates if this version is a release version
RELEASE = 'dev' not in VERSION

if not RELEASE:
    VERSION += get_git_devstr(False)

# Populate the dict of setup command overrides; this should be done before
# invoking any other functionality from distutils since it can potentially
# modify distutils' behavior.
cmdclassd = register_commands(PACKAGENAME, VERSION, RELEASE)

# Adjust the compiler in case the default on this platform is to use a
# broken one.
adjust_compiler(PACKAGENAME)

# Freeze build information in version.py
generate_version_py(PACKAGENAME, VERSION, RELEASE, get_debug_option())

# Treat everything in scripts except README.rst as a script to be installed
scripts = [fname for fname in glob.glob(os.path.join('scripts', '*'))
           if os.path.basename(fname) != 'README.rst']


try:

    from astropy.setup_helpers import get_package_info

    # Get configuration information from all of the various subpackages.
    # See the docstring for setup_helpers.update_package_files for more
    # details.
    package_info = get_package_info(PACKAGENAME)

    # Add the project-global data
    package_info['package_data'][PACKAGENAME] = ['data/*']

except ImportError: # compatibility with Astropy 0.2 - can be removed in cases
                    # where Astropy 0.2 is no longer supported

    from setuptools import find_packages
    from astropy.setup_helpers import filter_packages, update_package_files

    package_info = {}

    # Use the find_packages tool to locate all packages and modules
    package_info['packages'] = filter_packages(find_packages())

    # Additional C extensions that are not Cython-based should be added here.
    package_info['ext_modules'] = []

    # A dictionary to keep track of all package data to install
    package_info['package_data'] = {PACKAGENAME: ['data/*']}

    # A dictionary to keep track of extra packagedir mappings
    package_info['package_dir'] = {}

    # Update extensions, package_data, packagenames and package_dirs from
    # any sub-packages that define their own extension modules and package
    # data.  See the docstring for setup_helpers.update_package_files for
    # more details.
    update_package_files(PACKAGENAME, package_info['ext_modules'],
                         package_info['package_data'], package_info['packages'],
                         package_info['package_dir'])

# Include all .c files, recursively, including those generated by
# Cython, since we can not do this in MANIFEST.in with a "dynamic"
# directory name.
c_files = []
for root, dirs, files in os.walk(PACKAGENAME):
    for filename in files:
        if filename.endswith('.c'):
            c_files.append(
                os.path.join(
                    os.path.relpath(root, PACKAGENAME), filename))
package_info['package_data'][PACKAGENAME].extend(c_files)

setup(name=PACKAGENAME,
      version=VERSION,
      description=DESCRIPTION,
      scripts=scripts,
      requires=['astropy'],
      install_requires=['astropy'],
      provides=[PACKAGENAME],
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      license=LICENSE,
      url=URL,
      long_description=LONG_DESCRIPTION,
      cmdclass=cmdclassd,
      zip_safe=False,
      use_2to3=True,
      classifiers = [ 'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Astronomy',
        ],
      **package_info
)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def filternum(filt='foo'):
    """Return index number for SDSS filters either from a number or name.

    Parameters
    ----------
    filt : str
        The filter name.

    Returns
    -------
    filternum : int
        The corresponding filter number

    Examples
    --------
    >>> filternum('g')
    1
    """
    if filt == 'foo':
        return range(5)
    else:
        filters = { 'u':0, 'g':1, 'r':2, 'i':3, 'z':4 }
        return filters[filt]



# Autogenerated by Astropy-affiliated package pydl's setup.py on 2014-05-06 18:31:40.713347

from astropy.version_helpers import update_git_devstr, get_git_devstr

_last_generated_version = '0.2.2'

version = update_git_devstr(_last_generated_version)
githash = get_git_devstr(sha=True, show_warning=False)

major = 0
minor = 2
bugfix = 2

release = True
debug = False

try:
    from .utils._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"



#
#
#
def munu_to_radec(mu, nu, **kwargs):
    """Convert from SDSS great circle coordinates to equatorial coordinates.
    """
    import numpy as np
    from pydlutils.coord import stripe_to_eta
    from pydlutils.goddard.misc import cirrange
    if 'stripe' in kwargs:
        node = 95.0
        incl = stripe_to_incl(kwargs['stripe'])
    elif 'node' in kwargs and 'incl' in kwargs:
        node = kwargs['node']
        incl = kwargs['incl']
    else:
        raise ValueError('Must specify either STRIPE or NODE,INCL!')
    if mu.size != nu.size:
        raise ValueError('Number of elements in MU and NU must agree!')
    sinnu = np.sin(np.deg2rad(nu))
    cosnu = np.cos(np.deg2rad(nu))
    sini = np.sin(np.deg2rad(incl))
    cosi = np.cos(np.deg2rad(incl))
    sinmu = np.sin(np.deg2rad(mu-node))
    cosmu = np.cos(np.deg2rad(mu-node))
    xx = cosmu * cosnu
    yy = sinmu * cosnu * cosi - sinnu * sini
    zz = sinmu * cosnu * sini + sinnu * cosi
    ra = cirrange(np.rad2deg(np.arctan2(yy,xx)) + node)
    dec = np.rad2deg(np.arcsin(zz))
    if 'phi' in kwargs:
        phi = np.rad2deg(np.arctan2(cosmu * sini,
            (-sinmu * sinnu * sini + cosnu * cosi)*cosnu))
        return (ra,dec,phi)
    else:
        return (ra,dec)




#
#
#
def stripe_to_incl(stripe):
    """Convert from SDSS stripe number to an inclination relative to the equator.
    """
    from pydlutils.coord import stripe_to_eta
    dec_center = 32.5
    eta_center = stripe_to_eta(stripe)
    incl = eta_center + dec_center
    return incl




#
#
#
def stripe_to_eta(stripe):
    """Convert from SDSS great circle coordinates to equatorial coordinates.
    """
    stripe_sep = 2.5
    eta = stripe * stripe_sep - 57.5
    if stripe > 46:
        eta -= 180.0
    return eta




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
#
def file_lines(path,compress=False):
    """Replicates the IDL FILE_LINES() function.

    Given a path to a file name or a list of such paths, returns the number of
    lines in the file(s).

    Parameters
    ----------
    path : str or list of str
        Path to a file.  Can be a list of paths.
    compress : bool, optional
        If set to ``True``, assumes that all files in `path` are GZIP compressed.

    Returns
    -------
    file_lines : int or list of int
        The number of lines in `path`.  Returns a list of lengths if a list of
        files is supplied.

    Notes
    -----
    The ``/NOEXPAND_PATH`` option in IDL's FILE_LINES() is not implemented.

    References
    ----------
    http://www.exelisvis.com/docs/FILE_LINES.html

    Examples
    --------
    >>> from pydl import file_lines
    >>> from os.path import dirname, join
    >>> file_lines(join(dirname(__file__),'tests','t','this-file-contains-42-lines.txt'))
    42
    """
    scalar = False
    if isinstance(path, str) or isinstance(path, unicode):
        working_path = [path]
        scalar = True
    else:
        working_path = path
    lines = list()
    for filename in working_path:
        if compress:
            import gzip
            from sys import version_info
            if version_info[0] == 2 and version_info[1] < 7:
                f = gzip.open(filename)
                lines.append(len(f.readlines()))
                f.close()
            else:
                with gzip.open(filename) as f:
                    lines.append(len(f.readlines()))
        else:
            with open(filename) as f:
                lines.append(len(f.readlines()))
    if scalar:
        return lines[0]
    else:
        return lines



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def sdss_sweep_circle(ra,dec,radius,stype='star',allobj=False):
    """Read the SDSS datasweep files and return objects around a location.

    Parameters
    ----------
    ra, dec : float
        The sky location to search, J2000 degrees.
    radius : float
        The radius around `ra`, `dec` to search.
    stype : str, optional
        The type of object to search, 'star', 'gal' or 'sky'.  The default is 'star'.
    allobj : bool, optional
        If set to ``True``, return all objects found, not just SURVEY_PRIMARY.

    Returns
    -------
    sdss_sweep_circle : object
        The value of the bitmask name(s).

    Notes
    -----
    Assumes that the sweep files exist in ``$PHOTO_SWEEP`` and that index files
    have been created.
    """
    import numpy as np
    from os import getenv
    from os.path import join
    from astropy.io import fits
    from ... import uniq
    from .. import PydlutilsException
    from ..spheregroup import spherematch
    from . import sweep_cache, sdss_flagval
    #
    # Check values
    #
    if stype not in ('star','gal','sky'):
        raise ValueError('Invalid type {0}!'.format(stype))
    sweepdir = getenv('PHOTO_SWEEP')
    if sweepdir is None:
        raise PydlutilsException('PHOTO_SWEEP is not set!')
    #
    # Read the index
    #
    if sweep_cache[stype] is None:
        indexfile = join(sweepdir,'datasweep-index-{0}.fits'.format(stype))
        with fits.open(indexfile) as f:
            sweep_cache[stype] = f[1].data
    index = sweep_cache[stype]
    #
    # Match
    #
    ira = np.array([ra])
    idec = np.array([dec])
    m1,m2,d12 = spherematch(ira,idec,index['RA'],index['DEC'],radius+0.36,maxmatch=0)
    if len(m2) == 0:
        return None
    if not allobj:
        w = index['NPRIMARY'][m2] > 0
        if w.any():
            m2 = m2[w]
        else:
            return None
    #
    # Maximum number of objects
    #
    if allobj:
        n = index['IEND'][m2] - index['ISTART'][m2] + 1
        ntot = (np.where(n>0,n,np.zeros(n.shape,dtype=n.dtype))).sum()
    else:
        ntot = index['NPRIMARY'][m2].sum()
    #
    # Find unique run + camcol
    #
    rc = index['RUN'][m2]*6 + index['CAMCOL'][m2] - 1
    isort = rc.argsort()
    iuniq = uniq(rc[isort])
    istart = 0
    objs = None
    nobjs = 0
    for i in range(len(iuniq)):
        iend=iuniq[i]
        icurr=isort[istart:iend]
        #
        # Determine which file and range of rows
        #
        run = index['RUN'][m2[icurr[0]]]
        camcol = index['CAMCOL'][m2[icurr[0]]]
        rerun = index['RERUN'][m2[icurr[0]]]
        fields = index[m2[icurr]]
        ist = fields['ISTART'].min()
        ind = fields['IEND'].max()
        if ind >= ist:
            #
            # Read in the rows of that file
            #
            swfile = join(getenv('PHOTO_SWEEP'),rerun,
                'calibObj-{0:06d}-{1:1d}-{2}.fits.gz'.format(int(run),int(camcol),stype))
            with fits.open(swfile) as f:
                tmp_objs = f[1].data[ist:ind]
            if tmp_objs.size > 0:
                #
                # Keep only objects within the desired radius
                #
                tm1,tm2,d12 = spherematch(ira,idec,tmp_objs['RA'],tmp_objs['DEC'],radius,maxmatch=0)
                if len(tm2) > 0:
                    tmp_objs=tmp_objs[tm2]
                    #
                    # Keep only SURVEY_PRIMARY objects by default
                    #
                    if not allobj:
                        w = (tmp_objs['RESOLVE_STATUS'] & sdss_flagval('RESOLVE_STATUS','SURVEY_PRIMARY')) > 0
                        if w.any():
                            tmp_objs = tmp_objs[w]
                        else:
                            tmp_objs = None
                    if tmp_objs is not None:
                        if objs is None:
                            objs = np.zeros(ntot,dtype=tmp_objs.dtype)
                        objs[nobjs:nobjs+tmp_objs.size] = tmp_objs
                        nobjs += tmp_objs.size
        istart=iend+1
    if nobjs > 0:
        return objs[0:nobjs]
    else:
        return None



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
#
#
def get_juldate():
    """Returns the current Julian date.

    Uses the MJD trick & adds the offset to get JD.
    """
    import time
    mjd = time.time()/86400.0 + 40587.0
    return mjd + 2400000.5
#
#
#
def main(args=None):
    """Allow this module to be run in scripts.
    """
    jd = get_juldate()
    print(jd)
    return



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the cooling directory of idlutils.
"""
from .read_ds_cooling import read_ds_cooling



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
def spec_path(plate,path=None,topdir=None,run2d=None):
    """Return the path to spPlate files.

    Parameters
    ----------
    plate : int or ndarray
        The plate(s) to examine.
    path : str, optional
        If set, `path` becomes the full path for every plate. In other words,
        it completely short-circuits this function.
    topdir : str, optional
        Used to override the value of BOSS_SPECTRO_REDUX.
    run2d : str, optional
        Used to override the value of RUN2D.

    Returns
    -------
    spec_path : list
        A list of paths, one for each plate.
    """
    from os import getenv
    from os.path import join
    from numpy import array
    if isinstance(plate,int) or isinstance(plate,long) or plate.shape == ():
        platevec = array([plate],dtype='i4')
    else:
        platevec = plate
    if path is None:
        if topdir is None:
            topdir = getenv('BOSS_SPECTRO_REDUX')
        if run2d is None:
            run2d = getenv('RUN2D')
    paths = list()
    for p in platevec:
        if path is not None:
            paths.append(path)
        else:
            paths.append(join(topdir,run2d,str(p)))
    return paths




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
#
import numpy as np
from astropy.utils import lazyproperty
#
class pcomp(object):
    """Replicates the IDL PCOMP() function.

    Parameters
    ----------
    x : array_like
        A 2-D array with N rows and M columns.
    standardize : bool, optional
        If set to ``True``, the input data will have its mean subtracted off
        and will be scaled to unit variance.
    covariance : bool, optional.
        If set to ``True``, the covariance matrix of the data will be used for
        the computation.  Otherwise the correlation matrix will be used.

    Notes
    -----

    References
    ----------
    http://www.exelisvis.com/docs/PCOMP.html

    Examples
    --------
    """
    def __init__(self,x,standardize=False,covariance=False):
        from scipy.linalg import eigh
        if x.ndim != 2:
            raise ValueError('Input array must be two-dimensional')
        no,nv = x.shape
        self._nv = nv
        if standardize:
            xstd = x - np.tile(x.mean(0),no).reshape(x.shape)
            s = np.tile(xstd.std(0),no).reshape(x.shape)
            self._array = xstd/s
            self._xstd = xstd
        else:
            self._array = x
            self._xstd = None
        self._standardize = standardize
        if covariance:
            self._c = np.cov(self._array,rowvar=0)
        else:
            self._c = np.corrcoef(self._array,rowvar=0)
        self._covariance = covariance
        #
        # eigh is used for symmetric matrices
        #
        evals, evecs = eigh(self._c)
        #
        # Sort eigenvalues in descending order
        #
        ie = evals.argsort()[::-1]
        self._evals = evals[ie]
        self._evecs = evecs[:,ie]
        #
        # If necessary, add code to fix the signs of the eigenvectors.
        # http://www3.interscience.wiley.com/journal/117912150/abstract
        #
        return
    #
    #
    #
    @lazyproperty
    def coefficients(self):
        """The principal components.  These are the coefficients of `derived`.
        Basically, they are a re-scaling of the eigenvectors.
        """
        return self._evecs * np.tile(np.sqrt(self._evals),self._nv).reshape(self._nv,self._nv)
    #
    #
    #
    @lazyproperty
    def derived(self):
        """An N x M array containing the derived variables."""
        derived_data = np.dot(self._array,self.coefficients)
        if self._standardize:
            derived_data += self._xstd
        return derived_data
    #
    #
    #
    @lazyproperty
    def variance(self):
        """An array of the M variances of each derived variable."""
        return self._evals/self._c.trace()
    #
    #
    #
    @lazyproperty
    def eigenvalues(self):
        """An array of the M eigenvalues that correspond to the principal
        components.
        """
        return self._evals



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
def readspec(platein,fiber='all',**kwargs):
    """Read SDSS/BOSS spec2d & spec1d files.

    INPUTS:
        plate - Plate number(s)
    OPTIONAL INPUTS:
        topdir - Override the value of $BOSS_SPECTRO_REDUX.
        run2d  - Override the value of $RUN2D.
        run1d  - Override the value of $RUN1D
        path   - Override all path information with this directory name.
        align  - If set, align all the spectra in wavelength.
        znum   - If set, return the znum-th best fit reshift fit, instead of the best
    """
    import os
    import os.path
    from astropy.io import fits as pyfits
    import numpy as np
    from . import number_of_fibers, latest_mjd, spec_path, spec_append
    try:
        nplate = len(platein)
        plate = platein
    except TypeError:
        nplate = 1
        plate = np.array([platein],dtype='i4')
    if 'mjd' in kwargs:
        try:
            nmjd = len(kwargs['mjd'])
        except TypeError:
            nmjd = 1
        if nmjd != nplate:
            raise TypeError("Plate & MJD must have the same length!")
    if 'run2d' in kwargs:
        run2d = kwargs['run2d']
    else:
        run2d = os.getenv('RUN2D')
    if 'run1d' in kwargs:
        run1d = kwargs['run1d']
    else:
        run1d = os.getenv('RUN1D')
    if fiber == 'all':
        #
        # Read all fibers
        #
        nfibers = number_of_fibers(plate,**kwargs)
        total_fibers = nfibers.sum()
        platevec = np.zeros(total_fibers,dtype='i4')
        fibervec = np.zeros(total_fibers,dtype='i4')
        k = 0
        for p in np.unique(plate):
            n = np.unique(nfibers[plate == p])[0]
            platevec[k:k+n] = p
            fibervec[k:k+n] = np.arange(n) + 1
            k += n
    else:
        try:
            nfiber = len(fiber)
        except TypeError:
            nfiber = 1
        if nplate > 1 and nfiber > 1 and nplate != nfiber:
            raise TypeError("Plate & Fiber must have the same length!")
        if nplate > 1:
            platevec = np.array(plate,dtype='i4')
        else:
            platevec = np.zeros(nfiber,dtype='i4') + plate
        if nfiber > 1:
            fibervec = np.array(fiber,dtype='i4')
        else:
            fibervec = np.zeros(nplate,dtype='i4') + fiber
    if 'mjd' in kwargs:
        mjdvec = np.zeros(nplate,dtype='i4') + kwargs['mjd']
    else:
        mjdvec = latest_mjd(platevec,**kwargs)
    #
    # Now select unique plate-mjd combinations & read them
    #
    pmjd = ((np.array(platevec,dtype='u8') << 16) +
        np.array(mjdvec,dtype='u8'))
    #print(pmjd)
    upmjd = np.unique(pmjd)
    zupmjd = zip(upmjd>>16,upmjd&((1<<16)-1))
    #print(zupmjd)
    spplate_data = dict()
    hdunames = ('flux','invvar','andmask','ormask','disp','plugmap','sky','loglam',)
    for thisplate,thismjd in zupmjd:
        #thisplate = int(p>>16)
        #thismjd = int(np.bitwise_and(p,(1<<16)-1))
        pmjdindex = ((platevec==thisplate) & (mjdvec==thismjd)).nonzero()[0]
        thisfiber = fibervec[pmjdindex]
        #print(type(thisplate),type(thismjd))
        pmjdstr = "{0:04d}-{1:05d}".format(int(thisplate),int(thismjd))
        if 'path' in kwargs:
            sppath = [ kwargs['path'] ]
        else:
            sppath = spec_path(thisplate,run2d=run2d)
        spfile = os.path.join(sppath[0],"spPlate-{0}.fits".format(pmjdstr))
        print(spfile)
        spplate = pyfits.open(spfile)
        #
        # Get wavelength coefficients from primary header
        #
        npix = spplate[0].header['NAXIS1']
        c0 = spplate[0].header['COEFF0']
        c1 = spplate[0].header['COEFF1']
        coeff0 = np.zeros(thisfiber.size,dtype='d') + c0
        coeff1 = np.zeros(thisfiber.size,dtype='d') + c1
        loglam0 = c0 + c1*np.arange(npix,dtype='d')
        loglam = np.resize(loglam0,(thisfiber.size,npix))
        #
        # Read the data images
        #
        for k in range(len(hdunames)):
            try:
                tmp = spplate[k].data[thisfiber-1,:]
            except IndexError:
                tmp = loglam
            if hdunames[k] not in spplate_data:
                if k == 0:
                    allpmjdindex = pmjdindex
                    allcoeff0 = coeff0
                    allcoeff1 = coeff1
                #
                # Put the data into the return structure
                #
                if hdunames[k] == 'plugmap':
                    spplate_data['plugmap'] = dict()
                    for c in spplate[5].columns.names:
                        spplate_data['plugmap'][c] = tmp.field(c)
                else:
                    spplate_data[hdunames[k]] = tmp
            else:
                #
                # Append data
                #
                if k == 0:
                    allpmjdindex = np.concatenate((allpmjdindex,pmjdindex))
                    if 'align' in kwargs:
                        mincoeff0 = min(allcoeff0)
                        if mincoeff0 == 0 and coeff0[0] > 0:
                            allcoeff0 = coeff0[0]
                            allcoeff1 = coeff1[1]
                        if mincoeff0 > 0 and coeff0[0] == 0:
                            coeff0 = mincoeff0
                            coeff1 = allcoeff1[0]
                        ps = np.floor( (coeff0[0] - mincoeff0)/coeff1[0] + 0.5)
                        if ps > 0:
                            coeff0 = coeff0 - ps*coeff1
                        else:
                            allcoeff0 = allcoeff0 + ps*allcoeff1
                    else:
                        ps = 0
                    allcoeff0 = np.concatenate((allcoeff0,coeff0))
                    allcoeff1 = np.concatenate((allcoeff1,coeff1))
                if hdunames[k] == 'plugmap':
                    for c in spplate[5].columns.names:
                        spplate_data['plugmap'][c] = np.concatenate(
                            (spplate_data['plugmap'][c],tmp.field(c)))
                else:
                    spplate_data[hdunames[k]] = spec_append(spplate_data[hdunames[k]],tmp,pixshift=ps)
        spplate.close()
        #
        # Read photoPlate information, if available
        #
        photofile = os.path.join(sppath[0],"photoPlate-{0}.fits".format(pmjdstr))
        if not os.path.exists(photofile):
            #
            # Hmm, maybe this is an SDSS-I,II plate
            #
            photofile = os.path.join(os.getenv('SPECTRO_MATCH'),run2d,
                os.path.basename(os.getenv('PHOTO_RESOLVE')),"{0:04d}".format(int(thisplate)),
                "photoPlate-{0}.fits".format(pmjdstr))
        if os.path.exists(photofile):
            photop = pyfits.open(photofile)
            tmp = photop[1].data[thisfiber-1]
            if 'tsobj' not in spplate_data:
                spplate_data['tsobj'] = dict()
                for c in photop[1].columns.names:
                    spplate_data['tsobj'][c] = tmp.field(c)
            else:
                for c in photop[1].columns.names:
                    spplate_data['tsobj'][c] = np.concatenate(
                        (spplate_data['tsobj'][c],tmp.field(c)))
            photop.close()

        #
        # Read redshift information, if available.
        #
        if 'znum' in kwargs:
            zfile = os.path.join(sppath[0],run1d,"spZall-{0}.fits".format(pmjdstr))
        else:
            zfile = os.path.join(sppath[0],run1d,"spZbest-{0}.fits".format(pmjdstr))
        if os.path.exists(zfile):
            spz = pyfits.open(zfile)
            if 'znum' in kwargs:
                nper = spz[0].header['DIMS0']
                zfiber = (thisfiber-1)*nper + kwargs['znum'] - 1
            else:
                zfiber = thisfiber
            tmp = spz[1].data[zfiber-1]
            if 'zans' not in spplate_data:
                spplate_data['zans'] = dict()
                for c in spz[1].columns.names:
                    spplate_data['zans'][c] = tmp.field(c)
            else:
                for c in spz[1].columns.names:
                    spplate_data['zans'][c] = np.concatenate(
                        (spplate_data['zans'][c],tmp.field(c)))
            spz.close()
    #
    # Reorder the data.  At this point allpmjdindex is an index for which
    # fiber[allpmjdindex] == spplate['plugmap']['FIBERID'], so we have to
    # reverse this mapping.
    #
    j = allpmjdindex.argsort()
    for k in spplate_data:
        if isinstance(spplate_data[k],dict):
            for c in spplate_data[k]:
                if spplate_data[k][c].ndim == 2:
                    spplate_data[k][c] = spplate_data[k][c][j,:]
                else:
                    spplate_data[k][c] = spplate_data[k][c][j]
        else:
            spplate_data[k] = spplate_data[k][j,:]
    allcoeff0 = allcoeff0[j]
    allcoeff1 = allcoeff1[j]
    #
    # If necessary, recompute the wavelengths
    #
    nfibers,npixmax = spplate_data['flux'].shape
    if 'align' in kwargs:
        loglam0 = allcoeff0[0] + allcoeff1[1]*np.arange(npixmax,dtype='d')
        spplate_data['loglam'] = np.resize(loglam0,(nfibers,npixmax))
    return spplate_data




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def spec_append(spec1, spec2, pixshift=0):
    """Append the array spec2 to the array spec1 & return a new array.

    If the dimension of these arrays is the same, then append as [spec1,spec2].
    If not, increase the size of the smaller array & fill with zeros.

    Parameters
    ----------
    spec1, spec2 : ndarray
        Append `spec2` to `spec1`.
    pixshift : int, optional
        If `pixshift` is set to a positive integer, `spec2` will be padded with
        `pixshift` zeros on the left size.  If `pixshift` is set to a
        negative integer, `spec1` will be padded with abs(`pixshift`) zeros on
        the left side.  If not set, all zeros will be padded on the right side.

    Returns
    -------
    spec_append : ndarray
        A new array containing both `spec1` and `spec2`.
    """
    from numpy import zeros
    nrows1,npix1 = spec1.shape
    nrows2,npix2 = spec2.shape
    nrows = nrows1+nrows2
    nadd1 = 0
    nadd2 = 0
    if pixshift != 0:
        if pixshift < 0:
            nadd1 = -pixshift
        else:
            nadd2 = pixshift
    maxpix = max(npix1+nadd1,npix2+nadd2)
    spec3 = zeros((nrows,maxpix),dtype=spec1.dtype)
    spec3[0:nrows1,nadd1:nadd1+npix1] = spec1
    spec3[nrows1:nrows,nadd2:nadd2+npix2] = spec2
    return spec3
